#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import calendar
import datetime
import getpass
import json
import logging
import tempfile
from glob import glob
from typing import Optional, List

from ...params import KeeperParams
from ..importer import (
    BaseImporter, Record, Folder, RecordField, RecordReferences, SharedFolder, Permission, BaseDownloadMembership,
    replace_email_domain, FIELD_TYPE_ONE_TIME_CODE)
from .account import Account
from .exceptions import LastPassUnknownError
from .vault import Vault, TMPDIR_PREFIX
from . import fetcher


class LastPassImporter(BaseImporter):
    def __init__(self):
        super(LastPassImporter, self).__init__()

        self.vault = None
        self.addresses = []  # type: List[LastPassAddress]
        self.months = {}
        _months = ['', 'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August',
                   'September', 'October', 'November', 'December']
        for i in range(len(_months)):
            if _months[i]:
                month = _months[i].casefold()
                if month not in self.months:
                    self.months[month] = i

        for i in range(len(calendar.month_name)):
            if calendar.month_name[i]:
                month = calendar.month_name[i].casefold()
                if month not in self.months:
                    self.months[month] = i

    def card_expiration(self, from_lastpass):  # type: (str) -> str
        if from_lastpass:
            comp = [x.strip().casefold() for x in from_lastpass.split(',')]
            if len(comp) == 2 and all(comp):
                try:
                    year = int(comp[1])
                    if year < 200:
                        year += 2000
                        comp[1] = str(year)
                except ValueError:
                    pass
                if comp[0] in self.months:
                    return f'{self.months[comp[0]]:0>2}/{comp[1]}'
        return from_lastpass

    def lastpass_date(self, from_lastpass):  # type: (str) -> int
        if from_lastpass:
            comp = [x.strip().casefold() for x in from_lastpass.split(',')]
            if len(comp) == 3 and all(comp):
                try:
                    month = self.months[comp[0]]
                    day = int(comp[1])
                    year = int(comp[2])
                    dt = datetime.date(year, month, day)
                    return int(datetime.datetime.fromordinal(dt.toordinal()).timestamp() * 1000)
                except:
                    pass
        return -1

    def find_address(self, address):  # type: (LastPassAddress) -> Optional[int]
        for i in range(len(self.addresses)):
            if self.addresses[i] == address:
                return i + 1

    def append_address(self, address):  # type: (LastPassAddress) -> Optional[int]
        if isinstance(address, LastPassAddress):
            self.addresses.append(address)
            return len(self.addresses)

    def parse_typed_notes(self, notes):    # type: (str) -> dict
        lines = notes.split('\n')
        fields = {}
        key = ''
        value = ''
        for line in lines:
            k, s, v = line.partition(':')
            if s == ':':
                if key:
                    if key == 'Notes':
                        value += line
                    elif key == 'Private Key':
                        if k == 'Public Key':
                            fields[key] = value
                            key = k
                            value = v
                        else:
                            value += '\n' + line
                    else:
                        fields[key] = value
                        key = k
                        value = v
                else:
                    key = k
                    value = v
            else:
                if key:
                    value += '\n' + line
        if key:
            fields[key] = value
        return fields

    def cleanup(self):
        """Cleanup should be performed when finished with encrypted attachment files"""
        # Don't remove custom specified tmpdir
        if self.vault and self.tmpdir is None:
            self.vault.cleanup()

        old_tmpdir = glob(f'{tempfile.gettempdir()}/{TMPDIR_PREFIX}*')
        if len(old_tmpdir) > 0:
            old_tmpdirs_str = '\n'.join(old_tmpdir)
            warn_msg = f'Previous temporary directories from interrupted imports detected:\n{old_tmpdirs_str}'
            logging.warning(warn_msg)

    def do_import(self, name, users_only=False, old_domain=None, new_domain=None, tmpdir=None, **kwargs):
        dry_run = kwargs.get('dry_run') or False
        request_settings = {}
        params = kwargs.get('params')
        if isinstance(params, KeeperParams):
            request_settings['proxies'] = params.rest_context.proxies
            request_settings['certificate_check'] = params.rest_context.certificate_check
        if 'filter_folder' in kwargs and kwargs['filter_folder']:
            request_settings['filter_folder'] = kwargs['filter_folder']

        self.tmpdir = tmpdir
        username = name
        password = getpass.getpass(prompt='...' + 'LastPass Password'.rjust(30) + ': ', stream=None)
        print('Press <Enter> if account is not protected with Multifactor Authentication')
        twofa_code = getpass.getpass(prompt='...' + 'Multifactor Password'.rjust(30) + ': ', stream=None)
        if not twofa_code:
            twofa_code = None

        try:
            vault = Vault.open_remote(username, password, multifactor_password=twofa_code, users_only=users_only,
                                      tmpdir=tmpdir, get_attachments=not dry_run, **request_settings)
        except LastPassUnknownError as lpe:
            message = str(lpe)
            if message.startswith('Try again OR look for an email'):
                message += 'If you do not receive an email, go to LastPass > Account Settings > Advanced Settings ' \
                           'and ensure that "Disable Email Verification" is unchecked.'
            logging.warning(message)
            return
        else:
            self.vault = vault
            if len(vault.errors) > 0:
                err_list = '\n'.join(vault.errors)
                logging.warning(f'The following errors occurred retrieving Lastpass shared folder members:\n{err_list}')

        for shared_folder in vault.shared_folders:
            if shared_folder.name:
                folder = SharedFolder()
                folder.path = shared_folder.name

                folder.permissions = []
                if shared_folder.members:
                    for member in shared_folder.members:
                        perm = Permission()
                        perm.name = replace_email_domain(member['username'], old_domain, new_domain)
                        perm.manage_records = member['readonly'] == '0'
                        perm.manage_users = member['can_administer'] == '1'
                        folder.permissions.append(perm)
                if shared_folder.teams:
                    for team in shared_folder.teams:
                        perm = Permission()
                        perm.name = team['name']
                        perm.manage_records = team['readonly'] == '0'
                        perm.manage_users = team['can_administer'] == '1'
                        folder.permissions.append(perm)

                yield folder

        missing_titles = 0
        for account in vault.accounts:  # type: Account
            record = Record()
            is_secure_note = False
            if account.url:
                record.login_url = account.url.decode('utf-8', 'ignore')
                if record.login_url == 'http://sn':
                    is_secure_note = True
                    record.login_url = None
                elif record.login_url == 'http://group':
                    continue

            record.type = 'login'
            if account.id:
                record.uid = account.id
            if account.name:
                record.title = account.name.decode('utf-8', 'ignore')
            else:
                missing_titles += 1
                record.title = f'Missing Title {missing_titles}'
                logging.warning(f'Missing title in record from LastPass. Assigning title "{record.title}"')
            if account.username:
                record.login = account.username.decode('utf-8', 'ignore')
            if account.password:
                record.password = account.password.decode('utf-8', 'ignore')
            if account.totp_url:
                record.fields.append(
                    RecordField(type=FIELD_TYPE_ONE_TIME_CODE, value=account.totp_url)
                )
            if isinstance(account.last_modified, int) and account.last_modified > 0:
                record.last_modified = account.last_modified
            if isinstance(account.custom_fields, list):
                for cf in account.custom_fields:
                    field_label = cf.name
                    if cf.type == 'password':
                        field_type = 'secret'
                    elif cf.type == 'email':
                        field_type = 'email'
                    elif cf.type == 'textarea':
                        field_type = 'multiline'
                    elif cf.type == 'tel':
                        field_type = 'phone'
                    else:
                        field_type = 'text'
                    cf = RecordField(type=field_type, label=field_label, value=cf.value)
                    record.fields.append(cf)
            if len(account.attachments) > 0:
                if record.attachments is None:
                    record.attachments = []
                record.attachments = account.attachments
            if account.notes:
                try:
                    notes = account.notes.decode('utf-8', 'ignore')
                except UnicodeDecodeError:
                    notes = ''
                if notes.startswith('NoteType:'):
                    typed_values = self.parse_typed_notes(notes)
                    if 'NoteType' in typed_values:
                        note_type = typed_values.pop('NoteType', '')
                        notes = typed_values.pop('Notes', '')
                        typed_values.pop('Language', None)

                        if note_type == 'Bank Account':
                            self.populate_bank_account(record, typed_values)
                        elif note_type == 'Credit Card':
                            self.populate_credit_card(record, typed_values)
                        elif note_type == 'Address':
                            address = LastPassAddress.from_lastpass(typed_values)
                            if address:
                                addr_ref = self.append_address(address)
                                if addr_ref:
                                    record.uid = addr_ref
                                self.populate_address_only(record, address)
                                self.populate_address(record, typed_values)
                        elif note_type == 'Driver\'s License':
                            address_record = self.populate_driver_license(record, typed_values)
                            if address_record is not None:
                                yield address_record
                        elif note_type == 'Passport':
                            self.populate_passport(record, typed_values)
                        elif note_type == 'Social Security':
                            self.populate_ssn_card(record, typed_values)
                        elif note_type == 'Health Insurance' or note_type == 'Insurance':
                            self.populate_health_insurance(record, typed_values)
                        elif note_type == 'Membership':
                            self.populate_membership(record, typed_values)
                        elif note_type == 'Database':
                            self.populate_database(record, typed_values)
                        elif note_type == 'Server':
                            self.populate_server(record, typed_values)
                        elif note_type == 'SSH Key':
                            self.populate_ssh_key(record, typed_values)
                        elif note_type == 'Software License':
                            self.populate_software_license(record, typed_values)

                    username = typed_values.pop('Username', '')
                    if username:
                        if record.login:
                            if record.login != username:
                                cf = RecordField(label='Username', value=username)
                                if record.type:
                                    cf.type = 'login'
                                record.fields.append(cf)
                        else:
                            record.login = username

                    password = typed_values.pop('Password', '')
                    if password:
                        if record.password:
                            if record.password != password:
                                cf = RecordField(label='Password', value=password)
                                if record.type:
                                    cf.type = 'password'
                                record.fields.append(cf)
                        else:
                            record.password = password

                    url = typed_values.pop('URL', '')
                    if url:
                        if record.login_url:
                            if record.login_url != url:
                                cf = RecordField(label='URL', value=url)
                                if record.type:
                                    cf.type = 'url'
                                record.fields.append(cf)
                        else:
                            record.login_url = url

                    for key in typed_values:
                        value = typed_values[key]
                        if value:
                            if record.type:
                                cf = RecordField(type='text', label=key, value=str(value))
                            else:
                                cf = RecordField(label=key, value=str(value))
                            record.fields.append(cf)
                else:
                    if is_secure_note and not record.login and not record.password:
                        record.type = 'encryptedNotes'
                        cf = RecordField(type='note', value=notes)
                        record.fields.append(cf)
                        notes = ''
                    else:
                        if record.login_url == 'http://':
                            record.login_url = ''

                record.notes = notes

            if account.group or account.shared_folder:
                fol = Folder()
                if account.shared_folder:
                    fol.domain = account.shared_folder.name
                if account.group:
                    fol.path = account.group
                    if isinstance(fol.path, bytes):
                        fol.path = fol.path.decode('utf-8', 'ignore')
                record.folders = [fol]

            yield record

    def populate_address(self, record, notes):  # type: (Record, dict) -> None
        person = LastPassPersonName()
        person.first = notes.pop('First Name', '')
        person.middle = notes.pop('Middle Name', '')
        person.last = notes.pop('Last Name', '')

        if person.first or person.last:
            pf = RecordField(type='name')
            pf.value = {
                'first': person.first,
                'middle': person.middle,
                'last': person.last
            }
            record.fields.append(pf)

        dt = self.lastpass_date(notes.pop('Birthday', None))
        if dt != -1:
            dtf = RecordField(type='birthDate', value=dt)
            record.fields.append(dtf)

        email = notes.pop('Email Address', None)
        if email:
            dtf = RecordField(type='email', value=email)
            record.fields.append(dtf)
        for phone_type in ['Phone', 'Evening Phone', 'Mobile Phone', 'Fax']:
            phone = notes.pop(phone_type, '')
            if phone:
                try:
                    phone_dict = json.loads(phone)
                    if isinstance(phone_dict, dict):
                        if 'num' in phone_dict:
                            phone_number = phone_dict['num']
                            phone_ext = phone_dict.get('ext') or ''
                            phone_country_code = phone_dict.get('cc3l') or ''
                            phf = RecordField(type='phone', label=phone_type)
                            phf.value = {
                              #  'region': phone_country_code,
                                'number': phone_number,
                                'ext': phone_ext,
                                'type': ('Mobile' if phone_type.startswith('Mobile') else
                                         'Home' if phone_type.startswith('Evening') else
                                         'Work')
                            }
                            record.fields.append(phf)
                except:
                    pass

    def populate_address_only(self, record, lastpass_address):  # type: (Record, LastPassAddress) -> None
        if lastpass_address:
            record.type = 'address'
            address = RecordField(type='address')
            address.value = {
                'street1': lastpass_address.street1 or '',
                'street2': lastpass_address.street2 or '',
                'city': lastpass_address.city or '',
                'state': lastpass_address.state or '',
                'zip': lastpass_address.zip or '',
                'country': lastpass_address.country or '',
            }
            record.fields.append(address)

    def populate_credit_card(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'bankCard'
        card = RecordField(type='paymentCard')
        card.value = {
            'cardNumber': notes.pop('Number', ''),
            'cardExpirationDate': self.card_expiration(notes.pop('Expiration Date', '')),
            'cardSecurityCode': notes.pop('Security Code', '')
        }
        record.fields.append(card)
        card_holder = RecordField(type='text', label='cardholderName', value=notes.pop('Name on Card', ''))
        record.fields.append(card_holder)

        dt = self.lastpass_date(notes.pop('Start Date', None))
        if dt != -1:
            dtf = RecordField(type='date', label='Start Date', value=dt)
            record.fields.append(dtf)


    def populate_bank_account(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'bankAccount'
        bank = RecordField(type='bankAccount')
        bank.value = {
            'accountType': notes.pop('Account Type', ''),
            'routingNumber': notes.pop('Routing Number', ''),
            'accountNumber': notes.pop('Account Number', ''),
        }
        record.fields.append(bank)

    def populate_passport(self, record, notes): # type: (Record, dict) -> None
        record.type = 'passport'
        number = RecordField(type='accountNumber', label='passportNumber', value=notes.pop('Number', None))
        record.fields.append(number)
        person = LastPassPersonName.from_lastpass(notes.pop('Name', None))
        if person:
            pf = RecordField(type='name')
            pf.value = {
                'first': person.first,
                'middle': person.middle,
                'last': person.last
            }
            record.fields.append(pf)
        dt = self.lastpass_date(notes.pop('Date of Birth', None))
        if dt != -1:
            dtf = RecordField(type='birthDate', value=dt)
            record.fields.append(dtf)
        dt = self.lastpass_date(notes.pop('Expiration Date', None))
        if dt != -1:
            dtf = RecordField(type='expirationDate', value=dt)
            record.fields.append(dtf)
        dt = self.lastpass_date(notes.pop('Issued Date', None))
        if dt != -1:
            dtf = RecordField(type='date', label='dateIssued', value=dt)
            record.fields.append(dtf)

    def populate_driver_license(self, record, notes):  # type: (Record, dict) -> Optional[Record]
        record.type = 'driverLicense'
        account_number = RecordField(type='accountNumber', label='dlNumber', value=notes.pop('Number', ''))
        record.fields.append(account_number)
        dt = self.lastpass_date(notes.pop('Expiration Date', None))
        if dt != -1:
            dtf = RecordField(type='expirationDate', value=dt)
            record.fields.append(dtf)
        dt = self.lastpass_date(notes.pop('Date of Birth', None))
        if dt != -1:
            dtf = RecordField(type='birthDate', value=dt)
            record.fields.append(dtf)
        person = LastPassPersonName.from_lastpass(notes.pop('Name', None))
        if person:
            pf = RecordField(type='name')
            pf.value = {
                'first': person.first,
                'middle': person.middle,
                'last': person.last
            }
            record.fields.append(pf)
        address = LastPassAddress.from_lastpass(notes)
        address_record = None
        if address:
            ref_no = self.find_address(address)
            if ref_no:
                if record.references is None:
                    record.references = []
                address_ref = next((x for x in record.references if x.type == 'address'), None)
                if address_ref is None:
                    address_ref = RecordReferences(type='address')
                    record.references.append(address_ref)
                address_ref.uids.append(ref_no)
        return address_record

    def populate_ssn_card(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'ssnCard'
        number = RecordField(type='accountNumber', label='identityNumber', value=notes.pop('Number', None))
        record.fields.append(number)
        person = LastPassPersonName.from_lastpass(notes.pop('Name', None))
        if person:
            pf = RecordField(type='name')
            pf.value = {
                'first': person.first,
                'middle': person.middle,
                'last': person.last
            }
            record.fields.append(pf)

    def populate_health_insurance(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'healthInsurance'
        number = RecordField(type='accountNumber', value=notes.pop('Policy Number', None))
        record.fields.append(number)
        dt = self.lastpass_date(notes.pop('Expiration', None))
        if dt != -1:
            dtf = RecordField(type='date', label='Expiration', value=dt)
            record.fields.append(dtf)

    def populate_membership(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'membership'
        number = RecordField(type='accountNumber', value=notes.pop('Membership Number', None))
        record.fields.append(number)
        person = LastPassPersonName.from_lastpass(notes.pop('Member Name', None))
        if person:
            pf = RecordField(type='name')
            pf.value = {
                'first': person.first,
                'middle': person.middle,
                'last': person.last
            }
            record.fields.append(pf)
        dt = self.lastpass_date(notes.pop('Start Date', None))
        if dt != -1:
            dtf = RecordField(type='date', label='Start Date', value=dt)
            record.fields.append(dtf)
        dt = self.lastpass_date(notes.pop('Expiration Date', None))
        if dt != -1:
            dtf = RecordField(type='date', label='Expiration Date', value=dt)
            record.fields.append(dtf)

    def populate_database(self,  record, notes):  # type: (Record, dict) -> None
        record.type = 'databaseCredentials'
        db_type = RecordField(type='text', label='type', value=notes.pop('Type', None))
        record.fields.append(db_type)

        host = RecordField(type='host')
        host.value = {
            'hostName': notes.pop('Hostname', ''),
            'port': notes.pop('Port', ''),
        }
        record.fields.append(host)
        record.login_url = ''

    def populate_server(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'serverCredentials'
        host = RecordField(type='host')
        host.value = {
            'hostName': notes.pop('Hostname', ''),
            'port': notes.pop('Port', ''),
        }
        record.fields.append(host)

    def populate_ssh_key(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'sshKeys'
        passphrase = notes.pop('Passphrase', None)
        if passphrase:
            if record.password:
                if record.password != passphrase:
                    passphrase = RecordField(type='password', label='passphrase', value=passphrase)
                    record.fields.append(passphrase)
            else:
                record.password = passphrase
        host = RecordField(type='host')
        host.value = {
            'hostName': notes.pop('Hostname', ''),
            'port': notes.pop('Port', ''),
        }
        record.fields.append(host)
        private_key = notes.pop('Private Key', None)
        public_key = notes.pop('Public Key', None)
        if private_key or public_key:
            value = {
                'privateKey': private_key,
                'publicKey': public_key
            }
            pk = RecordField(type='keyPair', value=value)
            record.fields.append(pk)

        dt = self.lastpass_date(notes.pop('Date', None))
        if dt != -1:
            dtf = RecordField(type='date', value=dt)
            record.fields.append(dtf)

    def populate_software_license(self, record, notes):  # type: (Record, dict) -> None
        record.type = 'softwareLicense'
        number = RecordField(type='licenseNumber', value=notes.pop('License Key', None))
        record.fields.append(number)
        dt = self.lastpass_date(notes.pop('Purchase Date', None))
        if dt != -1:
            dtf = RecordField(type='date', label='dateActive', value=dt)
            record.fields.append(dtf)


class LastPassPersonName(object):
    def __init__(self):
        self.first = ''
        self.middle = ''
        self.last = ''

    @staticmethod
    def from_lastpass(name):  # type: (str) -> 'Optional[LastPassPersonName]'
        if not name:
            return None
        if not isinstance(name, str):
            return None
        person = LastPassPersonName()
        last, sep, other = name.partition(',')
        if sep == ',':
            person.last = last.strip()
            comps = [x for x in other.strip().split(' ') if x]
        else:
            comps = [x for x in name.split(' ') if x]
            person.last = comps.pop(-1)
        if len(comps) > 0:
            person.first = comps.pop(0)
        if len(comps) > 0:
            person.middle = ' '.join(comps)

        if not person.first and not person.last:
            return None

        return person


class LastPassAddress(object):
    def __init__(self):
        self.street1 = ''
        self.street2 = ''
        self.city = ''
        self.state = ''
        self.zip = ''
        self.country = ''

    @staticmethod
    def _compare_case_insensitive(s1, s2):  # type: (any, any) -> bool
        if isinstance(s1, str) and isinstance(s2, str):
            return s1.casefold() == s2.casefold()
        if s1 is None and s2 is None:
            return True
        return False

    def __eq__(self, other):
        if not isinstance(other, LastPassAddress):
            return False
        return (self._compare_case_insensitive(self.street1, other.street1) and
                self._compare_case_insensitive(self.street2, other.street2) and
                self._compare_case_insensitive(self.city, other.city) and
                self._compare_case_insensitive(self.state, other.state))

    @staticmethod
    def from_lastpass(notes):  # type: (dict) -> 'Optional[LastPassAddress]'
        if not isinstance(notes, dict):
            return None

        address = LastPassAddress()
        if 'Address 1' in notes:
            address.street1 = notes.pop('Address 1', '')
            address.street2 = notes.pop('Address 2', '')
            street3 = notes.pop('Address 3', '')
            if street3:
                address.street2 += f' {street3}'
                address.street2 = address.street2.strip()

        elif 'Address' in notes:
            s1, sep, s2 = notes.pop('Address', '').partition(',')
            address.street1 = s1.strip()
            if sep == ',':
                address.street2 = s2.strip()
        else:
            return None

        address.city = notes.pop('City / Town', '')
        address.state = notes.pop('State', '')
        address.zip = notes.pop('Zip / Postal Code', '')
        address.country = notes.pop('Country', '')

        return address


class LastpassMembershipDownload(BaseDownloadMembership):
    def download_membership(self, params, **kwargs):
        username = input('...' + 'LastPass Username'.rjust(30) + ': ')
        if not username:
            logging.warning('LastPass username is required')
            return
        password = getpass.getpass(prompt='...' + 'LastPass Password'.rjust(30) + ': ', stream=None)
        if not password:
            logging.warning('LastPass password is required')
            return

        print('Press <Enter> if account is not protected with Multifactor Authentication')
        twofa_code = getpass.getpass(prompt='...' + 'Multifactor Password'.rjust(30) + ': ', stream=None)
        if not twofa_code:
            twofa_code = None

        session = None
        request_settings = {
            'proxies': params.rest_context.proxies,
            'certificate_check': params.rest_context.certificate_check
        }
        try:
            session = fetcher.login(username, password, twofa_code, **request_settings)
            blob = fetcher.fetch(session, **request_settings)
            encryption_key = blob.encryption_key(username, password)
            vault = Vault(blob, encryption_key, session, shared_folder_details=False, get_attachments=False)

            lastpass_shared_folder = [x for x in vault.shared_folders]

            for lpsf in lastpass_shared_folder:
                logging.info('Loading shared folder membership for "%s"', lpsf.name)

                members, teams, error = fetcher.fetch_shared_folder_members(session, lpsf.id, **request_settings)
                sf = SharedFolder()
                sf.uid = lpsf.id
                sf.path = lpsf.name
                sf.permissions = []
                if members:
                    sf.permissions.extend((self._lastpass_permission(x) for x in members))
                if teams:
                    sf.permissions.extend((self._lastpass_permission(x, team=True) for x in teams))
                yield sf

        except Exception as e:
            logging.warning(e)
        finally:
            if session:
                fetcher.logout(session, **request_settings)

    @staticmethod
    def _lastpass_permission(lp_permission, team=False):  # type: (dict, Optional[bool]) -> Permission
        permission = Permission()
        if team:
            permission.name = lp_permission['name']
        else:
            permission.name = lp_permission['username']
        permission.manage_records = lp_permission['readonly'] == '0'
        permission.manage_users = lp_permission['can_administer'] == '1'
        return permission




#!/usr/bin/env python3

"""Code/Data shared between recordv3.py and recordv2.py."""

import urllib.parse


def display_totp_details(otp_url):
    """Display Time-based One Time Password details."""
    query_fields = extract_query_fields(otp_url)
    display_totp_query_fields(query_fields)


def extract_otp_url(data, is_v2):
    """Get the otpauth url from our decoded json as a string."""
    assert isinstance(data, dict)
    fields = data['fields']
    assert isinstance(fields, list)
    if is_v2:
        assert len(fields) == 1
        return fields[0]['data']
    # For v3 records
    for subdict in fields:
        if subdict['type'] == 'oneTimeCode':
            value = subdict['value']
            assert isinstance(value, list)
            assert len(value) == 1
            return value[0]
    raise AssertionError('No otpauth URL found')


def extract_query_fields(otp_url):
    """Get the query fields from an otpauth query string as a dict."""
    parsed_url = urllib.parse.urlparse(otp_url)
    query_string = parsed_url.query
    parsed_query = urllib.parse.parse_qs(query_string)
    return parsed_query


def display_totp_query_fields(query_fields):
    """Display the details of totp data."""
    for key in ('secret', 'issuer', 'period'):
        if key in query_fields:
            assert len(query_fields[key]) == 1
            print('{}: {}'.format(key, query_fields[key][0]))
        else:
            print('{} not found'.format(key))



import argparse
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from ..pam.router_helper import router_get_connected_gateways
from ...display import bcolors
from ...discovery_common.rule import Rules
from ...discovery_common.types import RuleTypeEnum


class PAMGatewayActionDiscoverRuleRemoveCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-rule-remove')
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name of UID')
    parser.add_argument('--rule-id', '-i', required=True, dest='rule_id', action='store',
                        help='Identifier for the rule')

    def get_parser(self):
        return PAMGatewayActionDiscoverRuleRemoveCommand.parser

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        gateway = kwargs.get("gateway")
        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f'{bcolors.FAIL}Discovery job gateway [{gateway}] was not found.{bcolors.ENDC}')
            return

        try:
            rule_id = kwargs.get("rule_id")
            rules = Rules(record=gateway_context.configuration, params=params)
            rule_item = rules.get_rule_item(rule_type=RuleTypeEnum.ACTION, rule_id=rule_id)
            if rule_item is None:
                raise ValueError("Rule Id does not exist.")
            rules.remove_rule(rule_item)

            print(f"{bcolors.OKGREEN}Rule has been removed.{bcolors.ENDC}")
        except Exception as err:
            print(f"{bcolors.FAIL}Rule was not removed: {err}{bcolors.ENDC}")



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import base64
import collections
import itertools
import json
import logging
import math
import os
import re
import requests
import time
from datetime import datetime
from functools import reduce
from urllib.parse import urlunparse, urlparse

from typing import Optional, Tuple, Iterable, List, Dict, Any, Union, Type

import google
from Cryptodome.PublicKey import RSA

from . import constants, rest_api, loginv3, utils, crypto, vault
from .display import bcolors
from .enterprise import query_enterprise as qe
from .error import KeeperApiError
from .params import KeeperParams, PublicKeys, LAST_RECORD_UID
from .proto import APIRequest_pb2, record_pb2, enterprise_pb2, router_pb2
from .proto.record_pb2 import GetShareObjectsRequest, GetShareObjectsResponse
from .record import Record
from .recordv3 import RecordV3
from .security_audit import attach_security_data, get_security_data_key_type
from .shared_folder import SharedFolder
from .subfolder import BaseFolderNode
from .sync_down import sync_down
from .team import Team
from .ttk import TTK

current_milli_time = lambda: int(round(time.time() * 1000))


# PKCS7 padding helpers
BS = 16
pad_binary = lambda s: s + ((BS - len(s) % BS) * chr(BS - len(s) % BS)).encode()
unpad_binary = lambda s: s[0:-s[-1]]
unpad_char = lambda s: s[0:-ord(s[-1])]

decode_uid_to_str = lambda uid: base64.urlsafe_b64encode(uid).decode().rstrip('=')

LOCALE = 'en_US'


def run_command(params, request):
    # type: (KeeperParams, dict) -> dict
    if 'client_version' not in request:
        request['client_version'] = rest_api.CLIENT_VERSION

    return rest_api.v2_execute(params.rest_context, request)


def login(params, new_login=False, login_ui=None):
    # type: (KeeperParams, bool, Optional[Any]) -> None

    logging.info('Logging in to Keeper Commander')
    flow = loginv3.LoginV3Flow(login_ui)
    try:
        flow.login(params, new_login=new_login)
    except loginv3.InvalidDeviceToken:
        logging.warning('Registering new device')
        flow.login(params, new_device=True)


def accept_account_transfer_consent(params):     # type: (KeeperParams) -> bool
    share_account_by = params.get_share_account_timestamp()
    print(constants.ACCOUNT_TRANSFER_MSG.format(share_account_by.strftime('%a, %b %d %Y')))

    expired = datetime.today() > share_account_by
    input_options = 'Accept/L(ogout)' if expired else 'Accept/L(ater)'
    answer = input('Do you accept Account Transfer policy? {}: '.format(input_options))
    answer = answer.lower()
    if answer.lower() == 'accept':
        ok = True
        requests = []
        if 'share_account_to' in params.settings:
            for role in params.settings['share_account_to']:
                request = {
                    'command': 'share_account',
                    'to_role_id': role['role_id'],
                }
                if not params.forbid_rsa and 'public_key' in role:
                    encoded_public = utils.base64_url_decode(role['public_key'])
                    public_key = crypto.load_rsa_public_key(encoded_public)
                    transfer_key = crypto.encrypt_rsa(params.data_key, public_key)
                    request['transfer_key'] = utils.base64_url_encode(transfer_key)

                requests.append(request)
        responses = execute_batch(params, requests)
        if isinstance(responses, list):
            for response in responses:
                if response['result_code'] != 'success':
                    logging.warning('Account Transfer policy acceptance error: %s',
                                    response.get('message') or response['result_code'])
                    ok = False
        if ok and params.forbid_rsa and params.enterprise_ec_key:
            try:
                share_data_key_with_enterprise(params)
            except Exception as e:
                logging.warning('Account Transfer policy acceptance error: %s', e)
                ok = False
        return ok
    else:
        return False


def share_data_key_with_enterprise(params):     # type: (KeeperParams) -> None
    rq = enterprise_pb2.EnterpriseUserDataKey()
    rq.userEncryptedDataKey = crypto.encrypt_ec(params.data_key, params.enterprise_ec_key)
    communicate_rest(params, rq, 'enterprise/set_enterprise_user_data_key')


def get_record_data_json_bytes(data):    # type: (dict) -> bytes
    """Get serialized and utf-8 encoded record data with padding"""
    data_str = json.dumps(data)
    padding = int(math.ceil(max(384, len(data_str)) / 16) * 16)
    if padding:
        data_str = data_str.ljust(padding)
    return data_str.encode('utf-8')


def pad_aes_gcm(json):
    # AES-GCM encryption leaks length of plaintext, so we pad the object prior to encryption.
    result = json
    json_bytes = json.encode('UTF-8') if isinstance(json, str) else json
    if isinstance(json_bytes, bytes):
        bytes_len = len(json_bytes)
        padded_len = max(384, bytes_len)
        # padded_len = math.ceil(padded_len / 16) * 16
        if padded_len % 16: padded_len = padded_len + 16 - (padded_len % 16)

        if padded_len != bytes_len:
            pad_len = abs(padded_len - bytes_len)
            pad = ' ' * pad_len # fill with spaces (byte value 0x20)
            result = result + pad if isinstance(result, str) else b''.join([result, pad.encode('UTF-8')])

    return result


def decrypt_rsa_key(encrypted_private_key, data_key):
    """ Decrypt the RSA private key
    PKCS1 formatted private key, which is described by the ASN.1 type:
    RSAPrivateKey ::= SEQUENCE {
          version           Version,
          modulus           INTEGER,  -- n
          publicExponent    INTEGER,  -- e
          privateExponent   INTEGER,  -- d
          prime1            INTEGER,  -- p
          prime2            INTEGER,  -- q
          exponent1         INTEGER,  -- d mod (p-1)
          exponent2         INTEGER,  -- d mod (q-1)
          coefficient       INTEGER,  -- (inverse of q) mod p
          otherPrimeInfos   OtherPrimeInfos OPTIONAL
    }
    """
    decoded_data = utils.base64_url_decode(encrypted_private_key)
    return RSA.importKey(crypto.decrypt_aes_v1(decoded_data, data_key))


def get_record(params, record_uid):
    """Return the referenced record cache"""
    record_uid = record_uid.strip()

    if not record_uid:
        logging.warning('No record UID provided')
        return

    if not params.record_cache:
        logging.warning('No record cache.  Sync down first.')
        return

    if record_uid not in params.record_cache:
        logging.warning('Record UID %s not found in cache.' % record_uid)
        return

    cached_rec = params.record_cache[record_uid]
    version = cached_rec.get('version', 2)

    try:
        rec = Record(record_uid)
        data = json.loads(cached_rec['data_unencrypted'])
        extra = json.loads(cached_rec['extra_unencrypted']) if 'extra_unencrypted' in cached_rec else None
        rec.load(data, version=version, revision=cached_rec['revision'], extra=extra)
        if not resolve_record_view_path(params, record_uid):
            rec.mask_password()
        return rec
    except:
        logging.error('**** Error decrypting record %s', record_uid)


def is_shared_folder(params,shared_folder_uid):
    shared_folder_uid = shared_folder_uid.strip()

    if not shared_folder_uid:
        return False

    if not params.shared_folder_cache:
        return False

    if shared_folder_uid not in params.shared_folder_cache:
        return False

    return True


def is_team(params,team_uid):
    team_uid = team_uid.strip()

    if not team_uid:
        return False

    if not params.team_cache:
        return False

    if team_uid not in params.team_cache:
        return False

    return True


def get_share_admins_for_shared_folder(params, shared_folder_uid):
    # type: (KeeperParams, str) -> Optional[List[str]]
    if params.enterprise_ec_key:
        if shared_folder_uid in params.shared_folder_cache:
            if 'share_admins' in params.shared_folder_cache[shared_folder_uid]:
                return params.shared_folder_cache[shared_folder_uid]['share_admins']

        try:
            rq = enterprise_pb2.GetSharingAdminsRequest()
            rq.sharedFolderUid = utils.base64_url_decode(shared_folder_uid)
            rs = communicate_rest(params, rq, 'enterprise/get_sharing_admins',
                                  rs_type=enterprise_pb2.GetSharingAdminsResponse)
            admins = [x.email for x in rs.userProfileExts if x.isShareAdminForSharedFolderOwner and x.isInSharedFolder]
        except Exception as e:
            logging.debug(e)
            return
        if shared_folder_uid in params.shared_folder_cache:
            params.shared_folder_cache[shared_folder_uid]['share_admins'] = admins

        return admins


def get_share_admins_for_record(params, record_uid):    # type: (KeeperParams, str) -> Optional[List[str]]
    if params.enterprise_ec_key:
        if record_uid in params.record_cache:
            if 'share_admins' in params.record_cache[record_uid]:
                return params.record_cache[record_uid]['share_admins']

        try:
            rq = enterprise_pb2.GetSharingAdminsRequest()
            rq.recordUid = utils.base64_url_decode(record_uid)
            rs = communicate_rest(params, rq, 'enterprise/get_sharing_admins',
                                  rs_type=enterprise_pb2.GetSharingAdminsResponse)
            admins = [x.email for x in rs.userProfileExts if x.isShareAdminForRequestedObject]
        except Exception as e:
            logging.debug(e)
            return
        if record_uid in params.record_cache:
            params.record_cache[record_uid]['share_admins'] = admins

        return admins


def get_shared_folder(params, shared_folder_uid):   # type: (KeeperParams, str) -> Optional[SharedFolder]
    """Return the referenced shared folder"""
    shared_folder_uid = shared_folder_uid.strip()

    if not shared_folder_uid:
        logging.warning('No shared folder UID provided')
        return None

    if not params.shared_folder_cache:
        logging.warning('No shared folder cache.  Sync down first.')
        return None

    if shared_folder_uid not in params.shared_folder_cache:
        logging.warning('Shared folder UID not found.')
        return None

    cached_sf = params.shared_folder_cache[shared_folder_uid]

    sf = SharedFolder(shared_folder_uid)
    sf.load(cached_sf, cached_sf['revision'])
    return sf

def get_share_objects(params):  # type: (KeeperParams) -> Dict[str, Dict[str, Any]]
    if not params.share_object_cache:
        rq = GetShareObjectsRequest()
        rs = communicate_rest(params, rq, 'vault/get_share_objects', rs_type=GetShareObjectsResponse)
        users_by_type = dict(
            relationship=rs.shareRelationships,
            family= rs.shareFamilyUsers,
            enterprise=rs.shareEnterpriseUsers,
            mc=rs.shareMCEnterpriseUsers,
        )
        get_users = lambda rs_data, cat: {su.username: dict(name=su.fullname, is_sa=su.isShareAdmin, enterprise_id=su.enterpriseId, status=su.status, category=cat) for su in rs_data}
        users = reduce(
            lambda a, b: {**a, **b},
            [get_users(users, cat) for cat, users in users_by_type.items()],
            {}
        )
        enterprises = {se.enterpriseId: se.enterprisename for se in rs.shareEnterpriseNames}
        get_teams = lambda rs_data: {utils.base64_url_encode(st.teamUid): dict(name=st.teamname, enterprise_id=st.enterpriseId) for st in rs_data}
        teams = get_teams(rs.shareTeams)
        teams_mc = get_teams(rs.shareMCTeams)

        share_objects = dict(
            users=users,
            enterprises=enterprises,
            teams={**teams, **teams_mc}
        )
        params.share_object_cache = share_objects
    return params.share_object_cache

def load_user_public_keys(params, emails, send_invites=False):  # type: (KeeperParams, List[str], bool) -> Optional[List[str]]
    s = set((x.casefold() for x in emails))
    s.difference_update(params.key_cache.keys())
    if not s:
        return

    emails_to_load = list(s)
    rq = APIRequest_pb2.GetPublicKeysRequest()
    rq.usernames.extend(emails_to_load)
    need_share_accept = []
    rs = communicate_rest(params, rq, 'vault/get_public_keys', rs_type=APIRequest_pb2.GetPublicKeysResponse)
    for pk in rs.keyResponses:
        email = pk.username
        if pk.errorCode in ['', 'success']:
            rsa = pk.publicKey
            ec = pk.publicEccKey
            aes = params.data_key if email == params.user else None
            params.key_cache[email] = PublicKeys(aes=aes, rsa=rsa, ec=ec)
        elif pk.errorCode == 'no_active_share_exist':
            need_share_accept.append(pk.username)
    if len(need_share_accept) > 0 and send_invites:
        for email in need_share_accept:
            rq = APIRequest_pb2.SendShareInviteRequest()
            rq.email = email
            try:
                communicate_rest(params, rq, 'vault/send_share_invite')
            except Exception as e:
                logging.debug('Share invite failed: %s', e)
        return need_share_accept


def load_team_keys(params, team_uids):          # type: (KeeperParams, List[str]) -> None
    s = set(team_uids)
    s.difference_update(params.key_cache.keys())
    if not s:
        return

    if params.enterprise:
        logging.debug('Resolve team keys shared with enterprise')
        tree_key = params.enterprise['unencrypted_tree_key']
        for t in params.enterprise.get('teams', []):
            team_uid = t.get('team_uid')
            if team_uid in s:
                try:
                    encrypted_team_key = t.get('encrypted_team_key')
                    if encrypted_team_key:
                        team_key = crypto.decrypt_aes_v2(utils.base64_url_decode(encrypted_team_key), tree_key)
                        params.key_cache[team_uid] = PublicKeys(aes=team_key)
                        s.remove(team_uid)
                except Exception as e:
                    logging.debug('Team UID \"%s\": Decrypt key error: %s', team_uid, str(e))

    if not s:
        return

    logging.debug('Loading %d team keys', len(s))
    uids_to_load = list(s)

    while len(uids_to_load) > 0:
        uids = uids_to_load[:90]
        uids_to_load = uids_to_load[90:]
        rq = {
            'command': 'team_get_keys',
            'teams': uids
        }
        rs = communicate(params, rq)
        if 'keys' in rs:
            for tk in rs['keys']:
                if 'key' in tk:
                    team_uid = tk['team_uid']
                    try:
                        aes = b''
                        rsa = b''
                        ec = b''
                        encrypted_key = utils.base64_url_decode(tk['key'])
                        key_type = tk['type']
                        if key_type == 1:
                            aes = crypto.decrypt_aes_v1(encrypted_key, params.data_key)
                        elif key_type == 2:
                            aes = crypto.decrypt_rsa(encrypted_key, params.rsa_key2)
                        elif key_type == 3:
                            aes = crypto.decrypt_aes_v2(encrypted_key, params.data_key)
                        elif key_type == 4:
                            aes = crypto.decrypt_ec(encrypted_key, params.ecc_key)
                        elif key_type == -1:
                            ec = encrypted_key
                        elif key_type == -3:
                            rsa = encrypted_key
                        params.key_cache[team_uid] = PublicKeys(rsa=rsa, aes=aes, ec=ec)
                    except Exception as e:
                        logging.debug(e)


def load_available_teams(params):
    if params.available_team_cache is not None:
        return

    rq = {
        'command': 'get_available_teams'
    }
    try:
        rs = communicate(params, rq)
        params.available_team_cache = rs.get('teams')
    except Exception as e:
        logging.debug(e)


def get_team(params,team_uid):
    """Return the referenced team """
    team_uid = team_uid.strip()

    if not team_uid:
        logging.warning('No team UID provided')
        return

    if not params.team_cache:
        logging.warning('No team cache.  Sync down first.')
        return

    if team_uid not in params.team_cache:
        logging.warning('Team UID not found.')
        return

    cached_team = params.team_cache[team_uid]
    team = Team(team_uid)
    team.load(cached_team)

    return team


def search_records(params, searchstring):
    """Search for string in record contents 
       and return array of Record objects """

    p = re.compile(searchstring.lower())
    search_results = []

    for record_uid in params.record_cache:
        rec = get_record(params, record_uid)
        if not rec:
            continue
        cached_rec = params.record_cache[record_uid] or {}

        if cached_rec.get('version') == 3:
            data = cached_rec.get('data_unencrypted')
            rec.record_type = RecordV3.get_record_type_name(data)
            target = RecordV3.values_to_lowerstring(data)
        else:
            target = rec.to_lowerstring()

        if p.search(target):
            search_results.append(rec)
            
    return search_results


def search_shared_folders(params, searchstring):
    """Search shared folders """

    p = re.compile(searchstring.lower())

    search_results = [] 

    for shared_folder_uid in params.shared_folder_cache:

        logging.debug('Getting Shared Folder UID: %s', shared_folder_uid)
        sf = get_shared_folder(params, shared_folder_uid)
        target = sf.to_lowerstring()

        if p.search(target):
            logging.debug('Search success')
            search_results.append(sf)
     
    return search_results


def search_teams(params, searchstring):
    """Search teams """

    p = re.compile(searchstring.lower())

    search_results = [] 

    for team_uid in params.team_cache:
        team = get_team(params, team_uid)

        target = team.to_lowerstring()

        if p.search(target):
            logging.debug('Search success')
            search_results.append(team)
     
    return search_results


def prepare_record(params, record):
    """ Prepares the Record() object to be sent to the Keeper Cloud API
        by serializing and encrypting it in the proper JSON format used for
        transmission.  If the record has no UID, one is generated and the
        encrypted record key is sent to the server.  If this record was
        converted from RSA to AES we send the new record key. If the record
        is in a shared folder, must send shared folder UID for edit permission.
    """

    record_object = {
        'version': 2,
        'client_modified_time': current_milli_time()
    }  # type: Dict[str, Any]

    if not record.record_uid:
        logging.debug('Generated Record UID: %s', record.record_uid)
        record.record_uid = generate_record_uid()

    record_object['record_uid'] = record.record_uid

    data = {}
    extra = {}
    udata = {}
    unencrypted_key = None
    if record.record_uid in params.record_cache:
        path = resolve_record_write_path(params, record.record_uid)
        if path:
            record_object.update(path)
        else:
            logging.error('You do not have edit permissions on this record')
            return None

        rec = params.record_cache[record.record_uid]

        data.update(json.loads(rec['data_unencrypted'].decode('utf-8')))
        if data.get('secret2') != record.password:
            params.queue_audit_event('record_password_change', record_uid=record.record_uid)

        if 'extra' in rec:
            extra.update(json.loads(rec['extra_unencrypted'].decode('utf-8')))
        if 'udata' in rec:
            udata.update(rec['udata'])
        unencrypted_key = rec['record_key_unencrypted']
        record_object['revision'] = rec['revision']
        if record.record_uid in params.meta_data_cache and params.meta_data_cache[record.record_uid].get('is_converted_record_type'):
            logging.debug('Converted record sends record key')
            record_object['record_key'] = \
                utils.base64_url_encode(crypto.encrypt_aes_v1(unencrypted_key, params.data_key))
    else:
        logging.debug('Generated record key')
        unencrypted_key = os.urandom(32)
        record_object['record_key'] = utils.base64_url_encode(crypto.encrypt_aes_v1(unencrypted_key, params.data_key))
        record_object['revision'] = 0

    data['title'] = record.title
    data['folder'] = record.folder
    data['secret1'] = record.login
    data['secret2'] = record.password
    data['link'] = record.login_url
    data['notes'] = record.notes
    data['custom'] = record.custom_fields
    Record.validate_record_data(data, extra, udata)

    record_object['data'] = \
        utils.base64_url_encode(crypto.encrypt_aes_v1(json.dumps(data).encode('utf-8'), unencrypted_key))
    record_object['extra'] = \
        utils.base64_url_encode(crypto.encrypt_aes_v1(json.dumps(extra).encode('utf-8'), unencrypted_key))
    record_object['udata'] = udata

    try:
        if params.license and 'account_type' in params.license:
            if record.password and params.license['account_type'] == 2:
                for record_uid in params.record_cache:
                    if record_uid != record.record_uid:
                        rec = get_record(params, record_uid)
                        if rec.password == record.password:
                            params.queue_audit_event('reused_password', record_uid=record.record_uid)
                            break
    except:
        pass

    return record_object


def prepare_record_v3(params, record):   # type: (KeeperParams, Record) -> Optional[Tuple[dict, Optional[bytes]]]
    """ Prepares the Record() object to be sent to the Keeper Cloud API
        by serializing and encrypting it in the proper JSON format used for
        transmission.  If the record has no UID, one is generated and the
        encrypted record key is sent to the server.  If the record is in a
        shared folder, must send shared folder UID for edit permission.
    """
    if not record.record_uid:
        logging.debug('Generated Record UID: %s', record.record_uid)
        record.record_uid = generate_record_uid()

    record_object = {
        'record_uid': record.record_uid,
        'client_modified_time': current_milli_time()
    }

    audit_data = None
    data = ''
    if record.record_uid in params.record_cache:
        path = resolve_record_write_path(params, record.record_uid)
        if path:
            record_object.update(path)
        else:
            share_admins = get_share_admins_for_record(params, record.record_uid)
            if not share_admins or params.user not in share_admins:
                logging.error('You do not have edit permissions on this record')
                return None
            record_object['record_uid'] = record.record_uid

        rec = params.record_cache[record.record_uid]

        data = rec['data_unencrypted'].decode('utf-8') if isinstance(rec['data_unencrypted'], bytes) else rec['data_unencrypted']
        try:
            if params.enterprise_ec_key:
                d = json.loads(data)
                rt_name = d.get('type') or ''
                fields = itertools.chain(d.get('fields') or [], (d.get('custom') or []))
                url_field = next((u.get('value') for u in fields if u.get('type') == 'url' and u.get('value')), None)
                url = ''
                if url_field and isinstance(url_field, list):
                    url = utils.url_strip(url_field[0])

                adata = {
                    'title': d.get('title', ''),
                    'record_type': rt_name,
                }
                if url:
                    adata['url'] = url  # url will only be supplied if there is one on the record
                audit_data = crypto.encrypt_ec(json.dumps(adata).encode('utf-8'), params.enterprise_ec_key)

        except Exception as e:
            logging.error(bcolors.FAIL + 'Invalid record type! Error: ' + str(e) + bcolors.ENDC)
            return None

        data = pad_aes_gcm(data)

        unencrypted_key = rec['record_key_unencrypted']
        record_object['revision'] = rec['revision']
    else:
        logging.debug('Generated record key')
        unencrypted_key = utils.generate_aes_key()
        key = crypto.decrypt_aes_v2(unencrypted_key, params.data_key)
        key = base64.urlsafe_b64encode(key)
        record_object['record_key'] = key
        record_object['revision'] = 0

    rdata = crypto.encrypt_aes_v2(bytes(data, 'utf-8'), unencrypted_key)
    record_object['data'] = base64.urlsafe_b64encode(rdata).decode('utf-8')

    try:
        if params.license and 'account_type' in params.license:
            if record.password and params.license['account_type'] == 2:
                for record_uid in params.record_cache:
                    if record_uid != record.record_uid:
                        rec = get_record(params, record_uid)
                        if rec.password == record.password:
                            params.queue_audit_event('reused_password', record_uid=record.record_uid)
                            break
    except:
        pass

    return record_object, audit_data


def execute_router_rest(params: KeeperParams, endpoint: str, payload: Optional[bytes]) -> Optional[bytes]:
    transmission_key = utils.generate_aes_key()
    encrypted_session_token = crypto.encrypt_aes_v2(utils.base64_url_decode(params.session_token), transmission_key)
    encrypted_transmission_key = rest_api.encrypt_with_keeper_key(params.rest_context, transmission_key)

    headers = {
        'TransmissionKey': base64.b64encode(encrypted_transmission_key).decode('ascii'),
        'Authorization': 'KeeperUser ' + base64.b64encode(encrypted_session_token).decode('ascii'),
    }

    if 'ROUTER_URL' in os.environ:
        up = urlparse(os.environ['ROUTER_URL'])
        url_comp = (up.scheme, up.netloc, f'api/user/{endpoint}', None, None, None)
    else:
        up = urlparse(params.rest_context.server_base)
        url_comp = ('https', f'connect.{up.hostname}', f'api/user/{endpoint}', None, None, None)
    url = urlunparse(url_comp)

    if payload is not None:
        payload = crypto.encrypt_aes_v2(payload, transmission_key)
    rs = requests.post(url, data=payload, headers=headers, proxies=params.rest_context.proxies,
                       verify=params.rest_context.certificate_check)
    if rs.status_code == 200:
        rs_body = rs.content
        if rs_body:
            router_response = router_pb2.RouterResponse()
            router_response.ParseFromString(rs_body)
            if router_response.responseCode == router_pb2.RouterResponseCode.RRC_OK:
                if router_response.encryptedPayload:
                    return crypto.decrypt_aes_v2(router_response.encryptedPayload, transmission_key)
            else:
                if router_response.responseCode == router_pb2.RouterResponseCode.RRC_BAD_REQUEST:
                    code = 'bad_request'
                elif router_response.responseCode == router_pb2.RouterResponseCode.RRC_NOT_ALLOWED:
                    code = 'not_allowed'
                else:
                    code = 'router_error'
                raise KeeperApiError(code, router_response.errorMessage)
    else:
        message = rs.reason
        raise KeeperApiError('router_error', f'{message}: {rs.status_code}')


def execute_router(params, endpoint, request, *, rs_type=None):
    logger = logging.getLogger('keeper.router')
    if logger.level <= logging.DEBUG:
        js = google.protobuf.json_format.MessageToJson(request) if request else ''
        logger.debug('>>> [ROUTER RQ] \"%s\": %s', endpoint, js)

    payload = request.SerializeToString() if request else None
    rs_bytes = execute_router_rest(params, endpoint, payload)
    if rs_type and rs_bytes:
        response = rs_type()
        response.ParseFromString(rs_bytes)
        if logger.level <= logging.DEBUG:
            js = google.protobuf.json_format.MessageToJson(response)
            logger.debug('>>> [ROUTER RS] \"%s\": %s', endpoint, js)
        return response
    return None


def execute_router_json(params, endpoint,  request):
    logger = utils.get_logger()
    payload: Optional[bytes] = None
    if isinstance(request, dict):
        js = json.dumps(request)
        payload = js.encode('utf-8')
        if logger.level <= logging.DEBUG:
            logger.debug('>>> [RQ] \"%s\": %s', endpoint, js)

    rs_bytes = execute_router_rest(params, endpoint, payload)
    if rs_bytes:
        response = json.loads(rs_bytes)
        if logger.level <= logging.DEBUG:
            logger.debug('>>> [RS] \"%s\": %s', endpoint, rs_bytes.decode('utf-8'))
        return response
    return None


def communicate_rest(params, request, endpoint, *, rs_type=None, payload_version=None):
    api_request_payload = APIRequest_pb2.ApiRequestPayload()
    if params.session_token:
        api_request_payload.encryptedSessionToken = utils.base64_url_decode(params.session_token)
    if request:
        if logging.getLogger().level <= logging.DEBUG:
            js = google.protobuf.json_format.MessageToJson(request)
            logging.debug('>>> [RQ] %s: %s', endpoint, js)
        api_request_payload.payload = request.SerializeToString()
    if isinstance(payload_version, int):
        api_request_payload.apiVersion = payload_version

    rs = rest_api.execute_rest(params.rest_context, endpoint, api_request_payload)
    if isinstance(rs, bytes):
        TTK.update_time_of_last_activity()
        if rs_type:
            proto_rs = rs_type()
            proto_rs.ParseFromString(rs)
            if logging.getLogger().level <= logging.DEBUG:
                js = google.protobuf.json_format.MessageToJson(proto_rs)
                logging.debug('>>> [RS] %s: %s', endpoint, js)
            return proto_rs
        else:
            return rs
    elif isinstance(rs, dict):
        kae = KeeperApiError(rs['error'], rs['message'])
        if kae.result_code == 'session_token_expired':
            params.session_token = None
        raise kae
    raise KeeperApiError('Error', endpoint)


def communicate(params, request, retry_on_throttle=True):
    # type: (KeeperParams, dict, Optional[bool]) -> dict

    request['client_time'] = current_milli_time()
    request['locale'] = LOCALE
    request['device_id'] = 'Commander'
    request['session_token'] = params.session_token
    request['username'] = params.user.lower()
    try:
        response_json = run_command(params, request)
        if response_json['result'] != 'success':
            if retry_on_throttle and response_json.get('result_code') == 'throttled':
                logging.info('Throttled. sleeping for 10 seconds')
                time.sleep(10)
                # Allow maximum 1 retry per call
                return communicate(params, request, retry_on_throttle=False)
            raise KeeperApiError(response_json['result_code'], response_json['message'])
        TTK.update_time_of_last_activity()
        return response_json
    except KeeperApiError as kae:
        if kae.result_code == 'session_token_expired':
            params.session_token = None
        raise kae


def execute_batch(params, requests):
    # type: (KeeperParams, List[dict]) -> List[dict]
    responses = []
    if not requests:
        return responses

    throttle_delay = 10
    chunk_size = 999
    queue = requests.copy()
    delay_next_batch = False

    while len(queue) > 0:
        chunk = queue[:chunk_size]
        queue = queue[chunk_size:]

        rq = {
            'command': 'execute',
            'requests': chunk
        }
        try:
            if delay_next_batch:
                time.sleep(throttle_delay)
            rs = communicate(params, rq)
            if 'results' in rs:
                results = rs['results']  # type: list
                if len(results) > 0:
                    error_rs = results[-1]
                    throttled = error_rs.get('result') != 'success' and error_rs.get('result_code') == 'throttled'
                    if throttled:
                        delay_next_batch = True
                        results.pop()
                responses.extend(results)

                if len(results) < len(chunk):
                    queue = chunk[len(results):] + queue
        except Exception as e:
            logging.error(e)

    return responses


def update_record(params, record, **kwargs):
    """ Push a record update to the cloud. 
        Takes a Record() object, converts to record JSON
        and pushes to the Keeper cloud API
    """

    if (record and record.record_uid in params.record_cache
            and 'version' in params.record_cache[record.record_uid]
            and params.record_cache[record.record_uid]['version'] == 3):
        return update_record_v3(params, record, **kwargs)

    record_rq = prepare_record(params, record)
    if record_rq is None:
        return

    request = {
        'command': 'record_update',
        'update_records': [record_rq]
    }
    response_json = communicate(params, request)

    new_revision = 0
    if 'update_records' in response_json:
        for info in response_json['update_records']:
            if info['record_uid'] == record.record_uid:
                if info['status'] == 'success':
                    new_revision = response_json['revision']

    if new_revision == 0:
        logging.error('Error: Revision not updated')
        return False

    if new_revision == record_rq['revision']:
        logging.error('Error: Revision did not change')
        return False

    if not kwargs.get('silent'):
        logging.info('Update record successful for record_uid=%s, revision=%d, new_revision=%s',
                     record_rq['record_uid'], record_rq['revision'], new_revision)

    record_rq['revision'] = new_revision

    # sync down the data which updates the caches
    sync_down(params)
    add_record_audit_data(params, [record.record_uid])
    return True


def get_pb2_record_update(params, rec, **kwargs):
    # type: (KeeperParams, Record, ...) -> Optional[dict]
    """Get an instance Protobuf RecordUpdate from an instance of Record (rec)

    Return a dictionary of necessary record items including the Protobuf RecordUpdate instance
    """
    rec_data = prepare_record_v3(params, rec)
    if rec_data is None:
        return
    record_rq, audit = rec_data

    links_by_uid = kwargs.get('record_links_by_uid')
    if links_by_uid:
        links_add_by_uid = links_by_uid.get('record_links_add') or {}
        links_add = links_add_by_uid.get(rec.record_uid) or []
        links_del_by_uid = links_by_uid.get('record_links_remove') or {}
        links_remove = links_del_by_uid.get(rec.record_uid) or []
    else:
        links = kwargs.get('record_links') or {}
        links_add = links.get('record_links_add') or []
        links_remove = links.get('record_links_remove') or []

    record_links_add = []
    for link in links_add:
        rl = record_pb2.RecordLink()
        rl.record_uid = link.get('record_uid')
        rl.record_key = link.get('record_key')
        if rl.record_uid and rl.record_key:
            record_links_add.append(rl)

    record_links_remove = [x['record_uid'] or b'' for x in links_remove]

    ru = record_pb2.RecordUpdate()
    uid = loginv3.CommonHelperMethods.url_safe_str_to_bytes(record_rq['record_uid'])
    data = loginv3.CommonHelperMethods.url_safe_str_to_bytes(record_rq['data'])
    ru.record_uid = uid
    ru.client_modified_time = record_rq['client_modified_time']
    ru.revision = record_rq['revision']
    ru.data = data
    #ru.non_shared_data = b''
    if record_links_add:
        ru.record_links_add.extend(record_links_add)
    if record_links_remove:
        ru.record_links_remove.extend(record_links_remove)
    if audit:
        ru.audit.data = audit
    ru = attach_security_data(params, rec.record_uid, ru)

    record_rq['pb2_record_update'] = ru
    return record_rq


def get_record_v3_response(params, rq, endpoint, record_rq_by_uid, silent=True):
    # type: (KeeperParams, Any, str, dict[str, dict], bool) -> Optional[bool]
    rs = communicate_rest(params, rq, endpoint)
    records_modify_rs = record_pb2.RecordsModifyResponse()
    records_modify_rs.ParseFromString(rs)

    for r in records_modify_rs.records:
        ruid = loginv3.CommonHelperMethods.bytes_to_url_safe_str(r.record_uid)
        record_rq = record_rq_by_uid[ruid]
        success = (r.status == record_pb2.RecordModifyResult.DESCRIPTOR.values_by_name['RS_SUCCESS'].number)
        status = record_pb2.RecordModifyResult.DESCRIPTOR.values_by_number[r.status].name

        if not success:
            logging.error(bcolors.FAIL + 'Error: Record update failed with status - %s' + bcolors.ENDC, status)
            return False

        new_revision = 0
        if success:
            new_revision = records_modify_rs.revision

        if new_revision == 0:
            logging.error('Error: Revision not updated')
            return False

        if new_revision == record_rq['revision']:
            logging.error('Error: Revision did not change')
            return False

        if not silent:
            logging.info(
                'Update record successful for record_uid=%s, revision=%d, new_revision=%s',
                ruid, record_rq['revision'], new_revision
            )

        record_rq['revision'] = new_revision

    params.sync_data = True
    return True


def update_record_v3(params, rec, **kwargs):   # type: (KeeperParams, Record, ...) -> Optional[bool]
    """ Push a record update to the cloud.
        Takes a Record() object, converts to record JSON
        and pushes to the Keeper cloud API
    """
    record_rq = get_pb2_record_update(params, rec, **kwargs)
    if not record_rq:
        return
    ru = record_rq['pb2_record_update']
    rq = get_records_update_request(params)
    rq.records.append(ru)
    record_rq_by_uid = {rec.record_uid: record_rq}
    return get_record_v3_response(params, rq, 'vault/records_update', record_rq_by_uid, silent=kwargs.get('silent'))


def update_records_v3(params, rec_list, **kwargs):   # type: (KeeperParams, List[Record], ...) -> Optional[bool]
    """ Push a record update to the cloud.
        Takes a Record() object, converts to record JSON
        and pushes to the Keeper cloud API
    """
    rq = get_records_update_request(params)
    record_rq_by_uid = {}
    for rec in rec_list:
        record_rq = get_pb2_record_update(params, rec, **kwargs)
        if record_rq:
            record_rq_by_uid[rec.record_uid] = record_rq
            rq.records.append(record_rq['pb2_record_update'])

    return get_record_v3_response(params, rq, 'vault/records_update', record_rq_by_uid)


def add_record(params, record,  **kwargs):   # type: (KeeperParams, Record, any) -> bool

    new_record = prepare_record(params, record)
    request = {
        'command': 'record_add',
        'record_uid': new_record['record_uid'],
        'record_type': 'password',
        'record_key': new_record['record_key'],
        'folder_type': 'user_folder',
        'how_long_ago': 0,
        'data': new_record['data'],
        'extra': new_record['extra'],
    }

    communicate(params, request)

    if not kwargs.get('silent'):
        logging.info('New record successful. record_uid=%s', new_record['record_uid'])

    # update record UID
    record.record_uid = new_record['record_uid']

    # sync down the data which updates the caches
    sync_down(params)
    params.environment_variables[LAST_RECORD_UID] = record.record_uid
    return True


def add_record_audit_data(params, record_uids):   # type: (KeeperParams, Iterable[str]) -> None
    if not params.enterprise_ec_key:
        return

    uids = set((x for x in record_uids if x in params.record_cache))
    audit_data = []   # type: List[record_pb2.RecordAddAuditData]
    for record_uid in uids:
        record = get_record(params, record_uid)
        if record:
            if record.title:
                title = record.title
                if len(title) > 900:
                    title = title[:900]
                audit = record_pb2.RecordAddAuditData()
                audit.record_uid = utils.base64_url_decode(record_uid)
                audit.revision = record.revision
                data = {
                    'title': title,
                    'record_type': record.record_type,
                }
                if record.login_url:
                    data['url'] = utils.url_strip(record.login_url)
                audit.data = crypto.encrypt_ec(json.dumps(data).encode('utf-8'), params.enterprise_ec_key)
                audit_data.append(audit)

    if audit_data:
        rq = record_pb2.AddAuditDataRequest()
        rq.records.extend(audit_data)
        try:
            communicate_rest(params, rq, 'vault/record_add_audit_data')
        except Exception as e:
            logging.info('Failed to store audit data: %s', str(e))


def add_record_v3(params, record, **kwargs):   # type: (KeeperParams, dict, ...) -> Optional[bool]
    """ Push a record update to the cloud.
        Takes a Record() object, converts to record JSON
        and pushes to the Keeper cloud API
    """

    record_rq = record
    if record_rq is None:
        return

    record_links = []
    links = kwargs.get('record_links') or {}
    links = links.get('record_links') or []

    for link in links:
        rl = record_pb2.RecordLink()
        rl.record_uid = link.get('record_uid')
        rl.record_key = link.get('record_key')
        if rl.record_uid and rl.record_key:
            record_links.append(rl)

    uid = loginv3.CommonHelperMethods.url_safe_str_to_bytes(record_rq['record_uid'])

    unencrypted_key = record_rq['record_key_unencrypted']
    key = crypto.encrypt_aes_v2(unencrypted_key, params.data_key)
    #key = base64.urlsafe_b64encode(key)

    data = record_rq['data_unencrypted'].decode('utf-8') if isinstance(record_rq['data_unencrypted'], bytes) else record_rq['data_unencrypted']
    try:
        d = json.loads(data)
        rt_name = d.get('type') or ''
        rt_def = RecordV3.resolve_record_type_by_name(params, rt_name)
        res = RecordV3.is_valid_record_type(data, rt_def)
        if not res.get('is_valid'):
            logging.info(res.get('error'))
    except Exception as e:
        logging.error(bcolors.FAIL + 'Invalid record type! Error: ' + str(e) + bcolors.ENDC)
        return None

    # Audit - if the title or url has changed and the user is an enterprise user
    audit = None
    if params.enterprise_ec_key:
        fields = itertools.chain(d.get('fields') or [], (d.get('custom') or []))
        url_field = next((u.get('value') for u in fields if u.get('type') == 'url' and u.get('value')), None)
        url = ''
        if url_field and isinstance(url_field, list):
            if len(url_field) > 0:
                url = utils.url_strip(str(url_field[0]))

        adata = {
            'title': d.get('title', ''),
            'record_type': rt_name,
        }
        if url: adata['url'] = url  # url will only be supplied if there is one on the record
        audit = record_pb2.RecordAudit()
        audit.version = 0
        audit.data = crypto.encrypt_ec(json.dumps(adata).encode('utf-8'), params.enterprise_ec_key)

    data = pad_aes_gcm(data)

    rdata = bytes(data, 'utf-8')
    rdata = crypto.encrypt_aes_v2(rdata, unencrypted_key)

    rq = kwargs.get('rq') or {}
    folder_type = rq.get('folder_type')
    folder_uid = rq.get('folder_uid')
    folder_key = rq.get('folder_key')
    if folder_type:
        folder_type_enum = {
            BaseFolderNode.RootFolderType: record_pb2.RecordFolderType.DESCRIPTOR.values_by_name['user_folder'].number,
            BaseFolderNode.UserFolderType: record_pb2.RecordFolderType.DESCRIPTOR.values_by_name['user_folder'].number,
            BaseFolderNode.SharedFolderType: record_pb2.RecordFolderType.DESCRIPTOR.values_by_name['shared_folder'].number,
            BaseFolderNode.SharedFolderFolderType: record_pb2.RecordFolderType.DESCRIPTOR.values_by_name['shared_folder_folder'].number
        }
        folder_type = folder_type_enum.get(folder_type)

    ra = record_pb2.RecordAdd()
    ra.record_uid = uid
    ra.record_key = key
    ra.client_modified_time = record_rq['client_modified_time']
    ra.data = rdata
    #ra.non_shared_data = b''
    if folder_uid and isinstance(folder_type, int):
        ra.folder_type = folder_type
        ra.folder_uid = loginv3.CommonHelperMethods.url_safe_str_to_bytes(folder_uid)
        if folder_key:
            ra.folder_key = folder_key
    if record_links:
        ra.record_links.extend(record_links)
    if audit:
        #ra.audit = audit # Assignment not allowed to field "audit" in protocol message object.
        ra.audit.version = audit.version
        ra.audit.data = audit.data
    ra = attach_security_data(params, record_rq, ra)

    rq = get_records_add_request(params)
    rq.records.append(ra)
    #NB! 'error code (Invalid data) has occurred.' won't return proper status - throws CommunicationError
    rs = communicate_rest(params, rq, 'vault/records_add')
    records_modify_rs = record_pb2.RecordsModifyResponse()
    records_modify_rs.ParseFromString(rs)

    for r in records_modify_rs.records:
        ruid = loginv3.CommonHelperMethods.bytes_to_url_safe_str(r.record_uid)
        success = (r.status == record_pb2.RecordModifyResult.DESCRIPTOR.values_by_name['RS_SUCCESS'].number)
        status = record_pb2.RecordModifyResult.DESCRIPTOR.values_by_number[r.status].name

        if not success:
            logging.error(bcolors.FAIL + 'Error: Record add failed with status - %s' + bcolors.ENDC, status)
            return False

        new_revision = 0
        if success and ruid == record['record_uid']:
            new_revision = records_modify_rs.revision

        if new_revision == 0:
            logging.error('Error: Revision not updated')
            return False

        if not kwargs.get('silent'):
            logging.debug('Add record successful for record_uid=%s, revision=%d', record_rq['record_uid'], new_revision)

        record_rq['revision'] = new_revision

    return True


def delete_record(params, record_uid):
    """ Delete a record """  
    request = {
        'command': 'record_update',
        'delete_records': [record_uid]
    }
    _ = communicate(params, request)
    logging.info('Record deleted with success')
    sync_down(params)
    return True


def generate_record_uid():
    """ Generate url safe base 64 16 byte uid """
    return base64.urlsafe_b64encode(os.urandom(16)).decode().rstrip('=')


def generate_aes_key():
    return os.urandom(32)


def resolve_record_permission_path(params, record_uid, permission):
    # type: (KeeperParams, str, str) -> Optional[Dict]

    for ap in enumerate_record_access_paths(params, record_uid):
        if ap.get(permission):
            path = {
                'record_uid': record_uid
            }
            if 'shared_folder_uid' in ap:
                path['shared_folder_uid'] = ap['shared_folder_uid']
            if 'team_uid' in ap:
                path['team_uid'] = ap['team_uid']
            return path

    return None


def resolve_record_write_path(params, record_uid):
    # type: (KeeperParams, str) -> dict or None
    return resolve_record_permission_path(params, record_uid, 'can_edit')


def resolve_record_share_path(params, record_uid):
    # type: (KeeperParams, str) -> dict or None
    return resolve_record_permission_path(params, record_uid, 'can_share')


def resolve_record_view_path(params, record_uid):
    # type: (KeeperParams, str) -> dict or None
    return resolve_record_permission_path(params, record_uid, 'can_view')


def resolve_record_access_path(params, record_uid, path=None):
    # type: (KeeperParams, str, dict or None) -> dict
    best_path = None

    for ap in enumerate_record_access_paths(params, record_uid):
        use_this_path = False
        if not best_path:
            use_this_path = True
        else:
            if not best_path.get('can_edit') and ap.get('can_edit'):
                use_this_path = True
            elif not best_path.get('can_share') and ap.get('can_share'):
                use_this_path = True
            elif not best_path.get('can_view') and ap.get('can_view'):
                use_this_path = True

        if use_this_path:
            best_path = ap
            if best_path.get('can_edit') and best_path.get('can_share') and best_path.get('can_view'):
                break

    if path is None:
        path = {
            'record_uid': record_uid
        }

    if best_path:
        if 'shared_folder_uid' in best_path:
            path['shared_folder_uid'] = best_path['shared_folder_uid']
        if 'team_uid' in best_path:
            path['team_uid'] = best_path['team_uid']

    return path


def enumerate_record_access_paths(params, record_uid):
    # type: (KeeperParams, str) -> collections.Iterable[dict]

    if record_uid in params.meta_data_cache:
        rmd = params.meta_data_cache[record_uid]
        yield {
            'record_uid': record_uid,
            'can_edit': rmd.get('can_edit') or False,
            'can_share': rmd.get('can_share') or False,
            'can_view': True
        }

    for sf_uid in params.shared_folder_cache:
        sf = params.shared_folder_cache[sf_uid]
        if 'records' in sf:
            sfrs = [x for x in sf['records'] if x['record_uid'] == record_uid]
            if len(sfrs) > 0:
                sfr = sfrs[0]
                is_owner = sfr.get('owner') is True
                if is_owner:
                    can_edit = True
                    can_share = True
                else:
                    can_edit = sfr['can_edit']
                    can_share = sfr['can_share']
                if 'key_type' in sf:
                    yield {
                        'record_uid': record_uid,
                        'shared_folder_uid': sf_uid,
                        'can_edit': can_edit,
                        'can_share': can_share,
                        'can_view': True
                    }
                else:
                    if 'teams' in sf:
                        for sf_team in sf['teams']:
                            team_uid = sf_team['team_uid']
                            if team_uid in params.team_cache:
                                team = params.team_cache[team_uid]
                                yield {
                                    'record_uid': record_uid,
                                    'shared_folder_uid': sf_uid,
                                    'team_uid': team_uid,
                                    'can_edit': can_edit and not team['restrict_edit'],
                                    'can_share': can_share and not team['restrict_share'],
                                    'can_view': not team['restrict_view']
                                }


def get_record_shares(params, record_uids, is_share_admin=False):
    # type: (KeeperParams, Iterable[str], bool) -> Optional[List[Dict[str, Any]]]
    def need_share_info(uid):
        if uid in params.record_cache:
            r = params.record_cache[uid]
            return 'shares' not in r
        return is_share_admin

    result = []
    unique = set(record_uids)
    uids = [x for x in unique if need_share_info(x)]
    try:
        while len(uids) > 0:
            chunk = uids[:999]
            uids = uids[999:]
            rq = record_pb2.GetRecordDataWithAccessInfoRequest()
            rq.clientTime = utils.current_milli_time()
            rq.recordUid.extend([utils.base64_url_decode(x) for x in chunk])
            rq.recordDetailsInclude = record_pb2.SHARE_ONLY
            rs = communicate_rest(params, rq, 'vault/get_records_details',
                                  rs_type=record_pb2.GetRecordDataWithAccessInfoResponse)
            for info in rs.recordDataWithAccessInfo:
                record_uid = utils.base64_url_encode(info.recordUid)
                rec = params.record_cache[record_uid] if record_uid in params.record_cache else {'record_uid': record_uid}  # type: dict
                if 'shares' not in rec:
                    rec['shares'] = {}
                rec['shares']['user_permissions'] = []
                rec['shares']['shared_folder_permissions'] = []
                for up in info.userPermission:
                    oup = {
                        'username': up.username,
                        'owner': up.owner,
                        'share_admin': up.shareAdmin,
                        'shareable': up.sharable,
                        'editable': up.editable,
                    }
                    if up.awaitingApproval:
                        oup['awaiting_approval'] = up.awaitingApproval
                    if up.expiration > 0:
                        oup['expiration'] = up.expiration
                    rec['shares']['user_permissions'].append(oup)
                for sp in info.sharedFolderPermission:
                    osp = {
                        'shared_folder_uid': utils.base64_url_encode(sp.sharedFolderUid),
                        'reshareable': sp.resharable,
                        'editable': sp.editable,
                        'revision': sp.revision,
                    }
                    if sp.expiration > 0:
                        osp['expiration'] = sp.expiration
                    rec['shares']['shared_folder_permissions'].append(osp)

                if record_uid not in params.record_cache:
                    result.append(rec)
    except Exception as e:
        logging.error(e)

    if len(result) > 0:
        return result


def query_enterprise(params, force=False, tree_key=None):
    try:
        if force is True and params.enterprise:
            params.enterprise = None
        qe(params, tree_key=tree_key)
    except Exception as e:
        share_account_by = params.get_share_account_timestamp()
        share_account_expired = share_account_by and datetime.today() > share_account_by
        # An exception is expected here if an Account Transfer is expired
        if share_account_expired:
            params.enterprise = None
        else:
            logging.warning(e, exc_info=True)


def login_and_get_mc_params_login_v3(params: KeeperParams, mc_id):

    resp = loginv3.LoginV3API.loginToMc(params.rest_context, params.session_token, mc_id)

    mc_params = KeeperParams(server=params.server)

    mc_params.config = params.config
    mc_params.debug = params.debug
    mc_params.auth_verifier = params.auth_verifier
    mc_params.salt = params.salt
    mc_params.iterations = params.iterations
    mc_params.user = params.user
    mc_params.account_uid_bytes = params.account_uid_bytes
    mc_params.password = params.password
    mc_params.enterprise_id = mc_id
    mc_params.session_token = params.session_token
    mc_params.data_key = params.data_key
    mc_params.rsa_key = params.rsa_key
    mc_params.rsa_key2 = params.rsa_key2
    mc_params.ecc_key = params.ecc_key
    mc_params.forbid_rsa = params.forbid_rsa

    mc_params.session_token = loginv3.CommonHelperMethods.bytes_to_url_safe_str(resp.encryptedSessionToken)
    mc_params.msp_tree_key = params.enterprise['unencrypted_tree_key']
    tree_key = crypto.decrypt_aes_v2(utils.base64_url_decode(resp.encryptedTreeKey), mc_params.msp_tree_key)

    sync_down(mc_params)
    query_enterprise(mc_params, True, tree_key=tree_key)

    return mc_params


def login_and_get_mc_params(params, mc_id):
    mc_params = KeeperParams(server=params.server)
    mc_params.config = params.config
    mc_params.auth_verifier = params.auth_verifier
    mc_params.salt = params.salt
    mc_params.iterations = params.iterations
    mc_params.user = params.user
    mc_params.password = params.password
    mc_params.account_uid_bytes = params.account_uid_bytes
    mc_params.enterprise_id = mc_id
    mc_params.msp_tree_key = params.enterprise['unencrypted_tree_key']
    mc_params.session_token = None

    login(mc_params)
    query_enterprise(mc_params)

    return mc_params


def get_correct_salt(salts):
    if len(salts) > 1:
        salt = next((s for s in salts if s.name.lower() == 'alternate'), salts[0])
    else:
        salt = salts[0]
    return salt


def send_keepalive(params):
    """Send a keepalive to the server, using protobufs."""
    communicate_rest(params, None, 'keep_alive')


def load_records_in_shared_folder(params, shared_folder_uid, record_uids=None):
    # type: (KeeperParams, str, Optional[Iterable[str]]) -> None

    if shared_folder_uid not in params.shared_folder_cache:
        raise Exception(f'Shared folder \"{shared_folder_uid}\" is not loaded.')
    shared_folder = params.shared_folder_cache[shared_folder_uid]
    shared_folder_key = shared_folder['shared_folder_key_unencrypted']
    record_keys = {}    # type: Dict[str, bytes]
    for rk in shared_folder.get('records', []):
        record_uid = rk.get('record_uid')
        try:
            key = utils.base64_url_decode(rk['record_key'])
            if len(key) == 60:
                record_key = crypto.decrypt_aes_v2(key, shared_folder_key)
            else:
                record_key = crypto.decrypt_aes_v1(key, shared_folder_key)
            record_keys[record_uid] = record_key
        except Exception as e:
            logging.debug('Cannot decrypt record \"%s\" key: %s', record_uid, e)

    if record_uids:
        record_set = set(record_uids)
        record_set.intersection_update(record_keys.keys())
    else:
        record_set = set(record_keys.keys())
    record_set.difference_update(params.record_cache.keys())

    while len(record_set) > 0:
        rq = record_pb2.GetRecordDataWithAccessInfoRequest()
        rq.clientTime = utils.current_milli_time()
        rq.recordDetailsInclude = record_pb2.DATA_PLUS_SHARE
        for uid in record_set:
            try:
                rq.recordUid.append(utils.base64_url_decode(uid))
            except Exception as e:
                logging.debug('Incorrect record UID \"%s\": %s', uid, e)
        record_set.clear()

        rs = communicate_rest(params, rq, 'vault/get_records_details', rs_type=record_pb2.GetRecordDataWithAccessInfoResponse)
        for record_info in rs.recordDataWithAccessInfo:
            record_uid = utils.base64_url_encode(record_info.recordUid)
            record_data = record_info.recordData
            try:
                if record_data.ownerRecordUid and record_data.encryptedLinkedRecordKey:
                    owner_id = utils.base64_url_encode(record_data.ownerRecordUid)
                    if owner_id in record_keys:
                        record_keys[record_uid] = crypto.decrypt_aes_v2(record_data.encryptedLinkedRecordKey, record_keys[owner_id])

                if record_uid not in record_keys:
                    continue

                record_key = record_keys[record_uid]
                version = record_data.version
                record = {
                    'record_uid': record_uid,
                    'revision': record_data.revision,
                    'version': version,
                    'shared': record_data.shared,
                    'data': record_data.encryptedRecordData,
                    'record_key_unencrypted': record_keys[record_uid],
                    'client_modified_time': record_data.clientModifiedTime,
                }
                data_decoded = utils.base64_url_decode(record_data.encryptedRecordData)
                if version <= 2:
                    record['data_unencrypted'] = crypto.decrypt_aes_v1(data_decoded, record_key)
                else:
                    record['data_unencrypted'] = crypto.decrypt_aes_v2(data_decoded, record_key)

                if record_data.encryptedExtraData and version <= 2:
                    record['extra'] = record_data.encryptedExtraData
                    extra_decoded = utils.base64_url_decode(record_data.encryptedExtraData)
                    record['extra_unencrypted'] = crypto.decrypt_aes_v1(extra_decoded, record_key)
                    record_data['udata'] = {
                        'file_id': []
                    }
                if version == 3:
                    v3_record = vault.KeeperRecord.load(params, record)
                    if isinstance(v3_record, vault.TypedRecord):
                        for ref in itertools.chain(v3_record.fields, v3_record.custom):
                            if ref.type.endswith('Ref') and isinstance(ref.value, list):
                                record_set.update(ref.value)
                elif version == 4:
                    if record_data.fileSize > 0:
                        record['file_size'] = record_data.fileSize
                    if record_data.thumbnailSize > 0:
                        record['thumbnail_size'] = record_data.thumbnailSize
                if record_data.ownerRecordUid and record_data.encryptedLinkedRecordKey:
                    record['owner_uid'] = utils.base64_url_encode(record_data.ownerRecordUid)
                    record['link_key'] = utils.base64_url_encode(record_data.encryptedLinkedRecordKey)

                record['shares'] = {
                    'user_permissions': [{
                        'username': up.username,
                        'owner': up.owner,
                        'share_admin': up.shareAdmin,
                        'shareable': up.sharable,
                        'editable': up.editable,
                        'awaiting_approval': up.awaitingApproval,
                        'expiration': up.expiration,
                    } for up in record_info.userPermission],
                    'shared_folder_permissions': [{
                        'shared_folder_uid': utils.base64_url_encode(sp.sharedFolderUid),
                        'reshareable': sp.resharable,
                        'editable': sp.editable,
                        'revision': sp.revision,
                        'expiration': sp.expiration,
                    } for sp in record_info.sharedFolderPermission],
                }

                params.record_cache[record_uid] = record
            except Exception as e:
                logging.debug('Error decrypting record \"%s\": %s', record_uid, e)
        record_set.difference_update(params.record_cache.keys())


def get_pb2_records_request(params, request_type='add'):
    # type: (KeeperParams, str) -> Union[record_pb2.RecordsUpdateRequest, record_pb2.RecordsAddRequest]
    rq_by_type = dict(add=record_pb2.RecordsAddRequest, update=record_pb2.RecordsUpdateRequest)
    new_rq_fn = rq_by_type.get(request_type.lower())
    if not new_rq_fn:
        raise Exception(f'Request type [{request_type} is unrecognized. Valid choices are [add, update]')
    else:
        rq = new_rq_fn()
        rq.client_time = utils.current_milli_time()
        rq.security_data_key_type = get_security_data_key_type(params)
        return rq


def get_records_add_request(params):   # type: (KeeperParams) -> record_pb2.RecordsAddRequest
    return get_pb2_records_request(params, 'add')


def get_records_update_request(params): # type: (KeeperParams) -> record_pb2.RecordsUpdateRequest
    return get_pb2_records_request(params, 'update')



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import logging

from .base import report_output_parser, Command, try_resolve_path, FolderMixin, dump_report_data, field_to_title
from ..error import CommandError
from .. import vault, generator, vault_extensions, utils

password_report_parser = argparse.ArgumentParser(prog='password-report', parents=[report_output_parser], description='Display record password report.')
password_report_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Display verbose information')
password_report_parser.add_argument('--policy', dest='policy', action='store',
                                    help='Password complexity policy. Length,Lower,Upper,Digits,Special. Default is 12,2,2,2,0')
password_report_parser.add_argument('-l', '--length', dest='length', type=int, action='store', help='Minimum password length.')
password_report_parser.add_argument('-u', '--upper', dest='upper', type=int, action='store', help='Minimum uppercase characters.')
password_report_parser.add_argument('--lower', dest='lower', type=int, action='store', help='Minimum lowercase characters.')
password_report_parser.add_argument('-d', '--digits', dest='digits', type=int, action='store', help='Minimum digits.')
password_report_parser.add_argument('-s', '--special', dest='special', type=int, action='store', help='Minimum special characters.')
password_report_parser.add_argument('folder', nargs='?', type=str, action='store', help='folder path or UID')


class PasswordReportCommand(Command):
    def get_parser(self):
        return password_report_parser

    def execute(self, params, **kwargs):
        verbose = kwargs.get('verbose') is True
        p_length = 0
        p_lower = 0
        p_upper = 0
        p_digits = 0
        p_special = 0

        policy = kwargs.get('policy')
        if policy:
            comps = [x.strip() for x in policy.split(',')]
            if any(False for c in comps if len(c) > 0 and not c.isdigit()):
                raise CommandError('Invalid policy format.  Must be list of integer values separated by commas.')
            if len(comps) > 0:
                if comps[0]:
                    p_length = int(comps[0])
            if len(comps) > 1:
                if comps[1]:
                    p_lower = int(comps[1])
            if len(comps) > 2:
                if comps[2]:
                    p_upper = int(comps[2])
            if len(comps) > 3:
                if comps[3]:
                    p_digits = int(comps[3])
            if len(comps) > 4:
                if comps[4]:
                    p_special = int(comps[4])
        else:
            if isinstance(kwargs.get('length'), int):
                p_length = kwargs.get('length')
            if isinstance(kwargs.get('upper'), int):
                p_upper = kwargs.get('upper')
            if isinstance(kwargs.get('lower'), int):
                p_lower = kwargs.get('lower')
            if isinstance(kwargs.get('digits'), int):
                p_digits = kwargs.get('digits')
            if isinstance(kwargs.get('special'), int):
                p_special = kwargs.get('special')

        if p_length <= 0 and p_upper <= 0 and p_lower <= 0 and p_digits <= 0 and p_special <= 0:
            raise CommandError('', 'Password policy must be specified.')

        path_or_uid = kwargs.get('folder')
        folder_uid = ''
        if path_or_uid:
            if path_or_uid in params.folder_cache:
                folder_uid = path_or_uid
            else:
                rs = try_resolve_path(params, path_or_uid)
                if rs is None:
                    raise CommandError('', f'Folder path {path_or_uid} not found')
                folder, pattern = rs
                if not folder or pattern:
                    raise CommandError('', f'Folder path {path_or_uid} not found')
                folder_uid = folder.uid or ''

        records = list(FolderMixin.get_records_in_folder_tree(params, folder_uid))
        table = []
        header = ['record_uid', 'title', 'description', 'length', 'lower', 'upper', 'digits', 'special']
        password_count = {}
        if verbose:
            header.append('score')
            if params.breach_watch:
                header.extend(['status', 'reused'])
                for record_uid, bw_record in params.breach_watch_records.items():
                    if record_uid in params.record_cache:
                        if isinstance(bw_record, dict):
                            data = bw_record.get('data_unencrypted')
                            if isinstance(data, dict):
                                passwords = data.get('passwords')
                                if isinstance(passwords, list):
                                    for pwd in passwords:
                                        password = pwd.get('value')
                                        if password:
                                            if password in password_count:
                                                password_count[password] += 1
                                            else:
                                                password_count[password] = 1

        fmt = kwargs.get('format')
        for record_uid in records:
            record = vault.KeeperRecord.load(params, record_uid)
            if not record:
                continue
            if record.version not in (2, 3):
                continue
            password = ''
            if isinstance(record, vault.PasswordRecord):
                password = record.password
            elif isinstance(record, vault.TypedRecord):
                password_field = record.get_typed_field('password')
                if password_field:
                    password = password_field.get_default_value(str)
            else:
                continue
            if not password:
                continue
            strength = generator.get_password_strength(password)
            password_ok = (strength.length >= p_length and strength.caps >= p_upper and strength.lower >= p_lower and
                           strength.digits >= p_digits and strength.symbols >= p_special)
            if password_ok:
                continue

            title = record.title
            if len(title) > 32:
                title = title[:30] + '...'
            description = vault_extensions.get_record_description(record)
            if isinstance(description, str):
                if len(description) > 32:
                    description = description[:30] + '...'
            row = [record_uid, title, description, strength.length, strength.lower, strength.caps, strength.digits, strength.symbols]
            if verbose:
                row.append(utils.password_score(password))
                if params.breach_watch:
                    status = ''
                    reused = None
                    bw_record = params.breach_watch_records.get(record_uid)
                    if isinstance(bw_record, dict):
                        data = bw_record.get('data_unencrypted')
                        if isinstance(data, dict):
                            passwords = data.get('passwords')
                            if isinstance(passwords, list):
                                password_status = next((x for x in passwords if x.get('value') == password), None)
                                if isinstance(password_status, dict):
                                    status = password_status.get('status')
                    if password in password_count:
                        reused_count = password_count[password]
                        if isinstance(reused_count, int) and reused_count > 1:
                            reused = reused_count
                    row.extend([status, reused])

            table.append(row)

        if fmt != 'json':
            header = [field_to_title(x) for x in header]

        logging.info('')
        if p_length > 0:
            logging.info('     Password Length: %d', p_length)
        if p_lower > 0:
            logging.info('Lowercase characters: %d', p_lower)
        if p_upper > 0:
            logging.info('Uppercase characters: %d', p_upper)
        if p_digits > 0:
            logging.info('              Digits: %d', p_digits)
        if p_special > 0:
            logging.info('  Special characters: %d', p_special)
        logging.info('')

        return dump_report_data(table, header, fmt=fmt, filename=kwargs.get('output'), row_number=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: breachwatch.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x62reachwatch.proto\x12\x0b\x42reachWatch\"\xa1\x01\n\x18\x42reachWatchRecordRequest\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x15\n\rencryptedData\x18\x02 \x01(\x0c\x12=\n\x13\x62reachWatchInfoType\x18\x03 \x01(\x0e\x32 .BreachWatch.BreachWatchInfoType\x12\x1c\n\x14updateUserWhoScanned\x18\x04 \x01(\x08\"z\n\x18\x42reachWatchUpdateRequest\x12G\n\x18\x62reachWatchRecordRequest\x18\x01 \x03(\x0b\x32%.BreachWatch.BreachWatchRecordRequest\x12\x15\n\rencryptedData\x18\x02 \x01(\x0c\"L\n\x17\x42reachWatchRecordStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x0e\n\x06reason\x18\x03 \x01(\t\"b\n\x19\x42reachWatchUpdateResponse\x12\x45\n\x17\x62reachWatchRecordStatus\x18\x01 \x03(\x0b\x32$.BreachWatch.BreachWatchRecordStatus\"3\n\x17\x42reachWatchTokenRequest\x12\x18\n\x10\x62reachWatchToken\x18\x01 \x01(\x0c\"M\n\x18\x42reachWatchTokenResponse\x12\x18\n\x10\x62reachWatchToken\x18\x01 \x01(\x0c\x12\x17\n\x0f\x63lientEncrypted\x18\x02 \x01(\x08\"Y\n\x17\x41nonymizedTokenResponse\x12\x13\n\x0b\x64omainToken\x18\x01 \x01(\x0c\x12\x12\n\nemailToken\x18\x02 \x01(\x0c\x12\x15\n\rpasswordToken\x18\x03 \x01(\x0c\"(\n\tHashCheck\x12\r\n\x05hash1\x18\x01 \x01(\x0c\x12\x0c\n\x04\x65uid\x18\x02 \x01(\x0c\"s\n\x18\x42reachWatchStatusRequest\x12\x17\n\x0f\x61nonymizedToken\x18\x01 \x01(\x0c\x12)\n\thashCheck\x18\x02 \x03(\x0b\x32\x16.BreachWatch.HashCheck\x12\x13\n\x0bremovedEuid\x18\x03 \x03(\x0c\"A\n\nHashStatus\x12\r\n\x05hash1\x18\x01 \x01(\x0c\x12\x0c\n\x04\x65uid\x18\x02 \x01(\x0c\x12\x16\n\x0e\x62reachDetected\x18\x03 \x01(\x08\"H\n\x19\x42reachWatchStatusResponse\x12+\n\nhashStatus\x18\x02 \x03(\x0b\x32\x17.BreachWatch.HashStatus\"Z\n\x1b\x45nterprisePublicKeyResponse\x12\x1b\n\x13\x65nterprisePublicKey\x18\x01 \x01(\x0c\x12\x1e\n\x16\x65nterpriseECCPublicKey\x18\x02 \x01(\x0c\"&\n\x0f\x46reeScanRequest\x12\x13\n\x0bhashedEmail\x18\x01 \x01(\x0c\"C\n\x10\x46reeScanResponse\x12\x15\n\remailBreaches\x18\x01 \x01(\x05\x12\x18\n\x10passwordBreaches\x18\x02 \x01(\x05\" \n\x0fPaidUserRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\"$\n\x10PaidUserResponse\x12\x10\n\x08paidUser\x18\x01 \x01(\x08\"$\n\x13\x44\x65tailedScanRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\"\'\n\x16UseOneTimeTokenRequest\x12\r\n\x05token\x18\x01 \x01(\x0c\"g\n\x0b\x42reachEvent\x12\x0c\n\x04site\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x18\n\x10passwordInBreach\x18\x03 \x01(\x08\x12\x0c\n\x04\x64\x61te\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\"\x89\x01\n\x17UseOneTimeTokenResponse\x12\x15\n\remailBreaches\x18\x01 \x01(\x05\x12\x18\n\x10passwordBreaches\x18\x02 \x01(\x05\x12.\n\x0c\x62reachEvents\x18\x03 \x03(\x0b\x32\x18.BreachWatch.BreachEvent\x12\r\n\x05\x65mail\x18\x04 \x01(\t\"-\n\x0fOneTimeUseToken\x12\r\n\x05\x65mail\x18\x01 \x01(\t\x12\x0b\n\x03pad\x18\x02 \x01(\t\"1\n\x17\x46reePasswordScanRequest\x12\x16\n\x0ehashedPassword\x18\x01 \x01(\x0c\"4\n\x18\x46reePasswordScanResponse\x12\x18\n\x10passwordBreaches\x18\x01 \x01(\x03*9\n\x13\x42reachWatchInfoType\x12\n\n\x06RECORD\x10\x00\x12\x16\n\x12\x41LTERNATE_PASSWORD\x10\x01\x42\'\n\x18\x63om.keepersecurity.protoB\x0b\x42reachWatchb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'breachwatch_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\013BreachWatch'
  _globals['_BREACHWATCHINFOTYPE']._serialized_start=1772
  _globals['_BREACHWATCHINFOTYPE']._serialized_end=1829
  _globals['_BREACHWATCHRECORDREQUEST']._serialized_start=35
  _globals['_BREACHWATCHRECORDREQUEST']._serialized_end=196
  _globals['_BREACHWATCHUPDATEREQUEST']._serialized_start=198
  _globals['_BREACHWATCHUPDATEREQUEST']._serialized_end=320
  _globals['_BREACHWATCHRECORDSTATUS']._serialized_start=322
  _globals['_BREACHWATCHRECORDSTATUS']._serialized_end=398
  _globals['_BREACHWATCHUPDATERESPONSE']._serialized_start=400
  _globals['_BREACHWATCHUPDATERESPONSE']._serialized_end=498
  _globals['_BREACHWATCHTOKENREQUEST']._serialized_start=500
  _globals['_BREACHWATCHTOKENREQUEST']._serialized_end=551
  _globals['_BREACHWATCHTOKENRESPONSE']._serialized_start=553
  _globals['_BREACHWATCHTOKENRESPONSE']._serialized_end=630
  _globals['_ANONYMIZEDTOKENRESPONSE']._serialized_start=632
  _globals['_ANONYMIZEDTOKENRESPONSE']._serialized_end=721
  _globals['_HASHCHECK']._serialized_start=723
  _globals['_HASHCHECK']._serialized_end=763
  _globals['_BREACHWATCHSTATUSREQUEST']._serialized_start=765
  _globals['_BREACHWATCHSTATUSREQUEST']._serialized_end=880
  _globals['_HASHSTATUS']._serialized_start=882
  _globals['_HASHSTATUS']._serialized_end=947
  _globals['_BREACHWATCHSTATUSRESPONSE']._serialized_start=949
  _globals['_BREACHWATCHSTATUSRESPONSE']._serialized_end=1021
  _globals['_ENTERPRISEPUBLICKEYRESPONSE']._serialized_start=1023
  _globals['_ENTERPRISEPUBLICKEYRESPONSE']._serialized_end=1113
  _globals['_FREESCANREQUEST']._serialized_start=1115
  _globals['_FREESCANREQUEST']._serialized_end=1153
  _globals['_FREESCANRESPONSE']._serialized_start=1155
  _globals['_FREESCANRESPONSE']._serialized_end=1222
  _globals['_PAIDUSERREQUEST']._serialized_start=1224
  _globals['_PAIDUSERREQUEST']._serialized_end=1256
  _globals['_PAIDUSERRESPONSE']._serialized_start=1258
  _globals['_PAIDUSERRESPONSE']._serialized_end=1294
  _globals['_DETAILEDSCANREQUEST']._serialized_start=1296
  _globals['_DETAILEDSCANREQUEST']._serialized_end=1332
  _globals['_USEONETIMETOKENREQUEST']._serialized_start=1334
  _globals['_USEONETIMETOKENREQUEST']._serialized_end=1373
  _globals['_BREACHEVENT']._serialized_start=1375
  _globals['_BREACHEVENT']._serialized_end=1478
  _globals['_USEONETIMETOKENRESPONSE']._serialized_start=1481
  _globals['_USEONETIMETOKENRESPONSE']._serialized_end=1618
  _globals['_ONETIMEUSETOKEN']._serialized_start=1620
  _globals['_ONETIMEUSETOKEN']._serialized_end=1665
  _globals['_FREEPASSWORDSCANREQUEST']._serialized_start=1667
  _globals['_FREEPASSWORDSCANREQUEST']._serialized_end=1716
  _globals['_FREEPASSWORDSCANRESPONSE']._serialized_start=1718
  _globals['_FREEPASSWORDSCANRESPONSE']._serialized_end=1770
# @@protoc_insertion_point(module_scope)



import argparse
import datetime
import itertools
import json
import logging
import operator
from functools import partial
from typing import Optional, Dict, Tuple, List, Any, Iterable, Union

from .base import GroupCommand, dump_report_data, field_to_title, report_output_parser
from .enterprise_common import EnterpriseCommand
from ..sox.sox_types import RecordPermissions
from .helpers.reporting import filter_rows
from .. import sox, api
from ..error import CommandError
from ..params import KeeperParams
from ..sox import sox_types, get_node_id
from ..sox.sox_data import SoxData

compliance_parser = argparse.ArgumentParser(add_help=False, parents=[report_output_parser])
rebuild_group = compliance_parser.add_mutually_exclusive_group()
rebuild_group.add_argument('--rebuild', '-r', action='store_true', help='rebuild local data from source')
nr_help = 'prevent remote data fetching if local cache present (invalid with --rebuild flag)'
rebuild_group.add_argument('--no-rebuild', '-nr', action='store_true', help=nr_help)
compliance_parser.add_argument('--no-cache', '-nc', action='store_true',
                               help='remove any local non-memory storage of data after report is generated')
compliance_parser.add_argument('--node', action='store', help='ID or name of node (defaults to root node)')
compliance_parser.add_argument('--regex', action='store_true', help='Allow use of regular expressions in search criteria')
compliance_parser.add_argument('pattern', type=str, nargs='*', help='Search string / pattern to filter results by. Multiple values allowed.')

default_report_parser = argparse.ArgumentParser(prog='compliance report', description='Run a compliance report.',
                                                parents=[compliance_parser])
username_opt_help = 'user(s) whose records are to be included in report (set option once per user)'
default_report_parser.add_argument('--username', '-u', action='append', help=username_opt_help)
job_title_opt_help = 'job title(s) of users whose records are to be included in report (set option once per job title)'
default_report_parser.add_argument('--job-title', '-jt', action='append', help=job_title_opt_help)
team_opt_help = 'name or UID of team(s) whose members\' records are to be included in report (set once per team)'
default_report_parser.add_argument('--team', action='append', help=team_opt_help)
record_search_help = 'UID or title of record(s) to include in report (set once per record). To allow non-exact ' \
                     'matching on record titles, include "*" where appropriate (e.g., to include records with titles' \
                     ' ending in "Login", set option value to "*Login")'
default_report_parser.add_argument('--record', action='append', help=record_search_help)
default_report_parser.add_argument('--url', action='append',
                                   help='URL of record(s) to include in report (set once per URL)')
default_report_parser.add_argument('--shared', action='store_true',
                                   help='show shared records only')
deleted_status_group = default_report_parser.add_mutually_exclusive_group()
deleted_status_group.add_argument('--deleted-items', action='store_true',
                                  help='show deleted records only (not valid with --active-items flag)')
deleted_status_group.add_argument('--active-items', action='store_true',
                                  help='show active records only (not valid with --deleted-items flag)')

team_report_desc = 'Run a report showing which shared folders enterprise teams have access to'
team_report_parser = argparse.ArgumentParser(prog='compliance team-report', description=team_report_desc,
                                             parents=[compliance_parser])
team_report_parser.add_argument('-tu', '--show-team-users', action='store_true', help='show all members of each team')

access_report_desc = 'Run a report showing all records a user has accessed or can access'
access_report_parser = argparse.ArgumentParser(prog='compliance record-access-report', description=access_report_desc,
                                               parents=[compliance_parser])
user_arg_help = 'username(s) or ID(s). Set once for each user to include. Set to "@all" to run report for all users'
access_report_parser.add_argument('--email', '-e', action='append', type=str, help=user_arg_help)
report_type_help = ('select type of record-access data to include in report (defaults to "history"). '
                    'Set to "history" to view past record-access activity, "vault" to view current vault contents')
ACCESS_REPORT_TYPES = ('history', 'vault')
access_report_parser.add_argument('--report-type', action='store', choices=ACCESS_REPORT_TYPES,
                                  default='history', help=report_type_help)
aging_help = 'include record-aging data (last modified, created, and last password rotation dates)'
access_report_parser.add_argument('--aging', action='store_true',  help=aging_help)

summary_report_desc = 'Run a summary compliance report'
summary_report_parser = argparse.ArgumentParser(prog='compliance summary-report', description=summary_report_desc,
                                                parents=[compliance_parser])
sf_report_desc = 'Run an enterprise-wide shared-folder report'
sf_report_parser = argparse.ArgumentParser(prog='compliance shared-folder-report', description=sf_report_desc,
                                           parents=[compliance_parser])
sf_report_parser.add_argument('-tu', '--show-team-users', action='store_true', help='show all members of each team')


def register_commands(commands):
    commands['compliance'] = ComplianceCommand()


def register_command_info(aliases, command_info):
    aliases['cr'] = ('compliance', 'report')
    aliases['compliance-report'] = ('compliance', 'report')
    command_info['compliance'] = 'Compliance Reporting for auditing'


def get_email(sdata, user_uid):    # type: (SoxData, int) -> str
    return sdata.get_users().get(user_uid).email


def get_team_usernames(sdata, team):  # type: (SoxData, sox_types.Team) -> List[str]
    return [get_email(sdata, userid) for userid in team.users]


class ComplianceCommand(GroupCommand):
    def __init__(self):
        super(ComplianceCommand, self).__init__()
        self.register_command('report', ComplianceReportCommand(), 'Run default compliance report')
        self.register_command('team-report', ComplianceTeamReportCommand(), team_report_desc, 'tr')
        self.register_command('record-access-report', ComplianceRecordAccessReportCommand(), access_report_desc, 'rar')
        self.register_command('summary-report', ComplianceSummaryReportCommand(), summary_report_desc, 'stats')
        self.register_command('shared-folder-report', ComplianceSharedFolderReportCommand(), sf_report_desc, 'sfr')
        self.default_verb = 'report'

    def validate(self, params):  # type: (KeeperParams) -> None
        sox.validate_data_access(params, cmd='compliance')


class BaseComplianceReportCommand(EnterpriseCommand):
    def __init__(self, report_headers, allow_no_opts=True, prelim_only=False):
        super(BaseComplianceReportCommand, self).__init__()
        self.title = None
        self.report_headers = report_headers
        self.allow_no_opts = allow_no_opts
        self.prelim_only = prelim_only
        self.group_by_column = None

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        pass

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        return []

    def show_help_text(self, local_data):
        pass

    def execute(self, params, **kwargs):  # type: (KeeperParams, any) -> any
        node_name_or_id = kwargs.get('node')
        node_id = get_node_id(params, node_name_or_id)
        enterprise_id = node_id >> 32
        nodes = params.enterprise['nodes']
        root_node_id = nodes[0].get('node_id', 0)
        max_data_age = datetime.timedelta(days=1)
        min_data_ts = 0 if kwargs.get('no_rebuild') else (datetime.datetime.now() - max_data_age).timestamp()

        default_opts = {'command', 'action', 'format'}
        opts_set = [val for opt, val in kwargs.items() if val and opt not in default_opts]
        if not opts_set and not self.allow_no_opts:
            local_sox_data = sox.get_compliance_data(params, node_id, enterprise_id, False, min_updated=0)
            self.show_help_text(local_sox_data)
            return

        rebuild = kwargs.get('rebuild')
        no_cache = kwargs.get('no_cache')
        shared_only = kwargs.get('shared')
        get_sox_data_fn = sox.get_prelim_data if self.prelim_only else sox.get_compliance_data
        fn_args = [params, enterprise_id] if self.prelim_only else [params, node_id, enterprise_id]
        fn_kwargs = {'rebuild': rebuild, 'min_updated': min_data_ts, 'no_cache': no_cache, 'shared_only': shared_only}
        sd = get_sox_data_fn(*fn_args, **fn_kwargs)
        report_fmt = kwargs.get('format', 'table')
        report_data = self.generate_report_data(params, kwargs, sd, report_fmt, node_id, root_node_id)
        patterns = kwargs.get('pattern', [])
        if '@all' in patterns:
            patterns.remove('@all')
        if patterns:
            report_data = filter_rows(report_data, patterns, use_regex=kwargs.get('regex'))
        headers = self.report_headers if report_fmt == 'json' else [field_to_title(h) for h in self.report_headers]
        report = dump_report_data(report_data, headers, title=self.title, fmt=report_fmt, filename=kwargs.get('output'),
                                  column_width=32, group_by=self.group_by_column)
        return report


class ComplianceReportCommand(BaseComplianceReportCommand):
    def __init__(self):
        headers = ['record_uid', 'title', 'type', 'username', 'permissions', 'url', 'in_trash', 'shared_folder_uid']
        super(ComplianceReportCommand, self).__init__(headers, allow_no_opts=False)

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return default_report_parser

    def show_help_text(self, local_data):  # type: (sox.sox_data.SoxData) -> None
        last_update_ts = local_data.storage.last_compliance_data_update
        if not last_update_ts:
            logging.info("Cache last update: NONE -- to build the cache, call the following command:"
                         "\n\t'compliance-report --rebuild'")
        else:
            last_update = datetime.datetime.fromtimestamp(last_update_ts)
            num_users = len(local_data.get_users())
            num_shared = len([rec for rec in local_data.get_records().values() if rec.shared])
            msg = f'cache last update: {last_update} Total records: {local_data.record_count} ' \
                  f'Total Users: {num_users} Total shared records: {num_shared}'
            logging.info(msg)
        help_txt = "\nGet record and sharing information from all vaults in the enterprise\n" \
                   "Format:\ncompliance-report [-h] [--rebuild] [--no-cache] [--node NODE] [--username USERNAME] " \
                   "[--job-title JOB_TITLE] [--team TEAM] [--record RECORD] [--url DOMAIN] [--shared] " \
                   "[--format {table,csv,json,pdf}] [--output OUTPUT] " \
                   "\n\nExamples:" \
                   "\nSee all records for a user" \
                   "\n\t'compliance-report --username USERNAME'" \
                   "\nFind all records for a specific URL that have been shared" \
                   "\n\t'compliance-report --shared --url URL'" \
                   "\nExport report of all shared records in a node" \
                   "\n\t'compliance-report --shared --output compliance_share_report.csv --format csv " \
                   "--node NODE_NAME_OR_ID'" \
                   "\n\t* use 'enterprise-info --node' to see a list of available nodes" \
                   "\nCache controls" \
                   "\n\t'compliance-report --rebuild'  " \
                   "\tUpdate and rebuild the cache with new compliance information." \
                   "\n\t'compliance-report --no-cache'\tDelete the cache entirely.The cache will be rebuilt if " \
                   "compliance-report is run again."
        logging.info(help_txt)

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        # type: (KeeperParams, Dict[str, Any], SoxData, str, int, int) -> List[List[Union[str, Any]]]
        def filter_owners(rec_owners):
            def filter_by_teams(users, teams):
                enterprise_teams = params.enterprise.get('teams', [])
                team_uids = {t.get('team_uid') for t in enterprise_teams}
                enterprise_team_users = params.enterprise.get('team_users', [])

                def get_team_users(team_ref):
                    team_ids = {team_ref} if team_ref in team_uids \
                        else {t.get('team_uid') for t in enterprise_teams if team_ref == t.get('name')}
                    return {u.get('enterprise_user_id') for u in enterprise_team_users if u.get('team_uid') in team_ids}

                team_users = set()
                for t_ref in teams:
                    team_users.update(get_team_users(t_ref))

                return [u for u in users if u.user_uid in team_users]

            usernames = kwargs.get('username')
            filtered = [o for o in rec_owners if o.email in usernames] if usernames else rec_owners
            job_titles = kwargs.get('job_title')
            filtered = [o for o in filtered if o.job_title in job_titles] if job_titles else filtered
            filtered = [o for o in filtered if o.node_id == node] if node != root_node else filtered
            team_refs = kwargs.get('team')
            filtered = filter_by_teams(filtered, team_refs) if team_refs else filtered
            return filtered

        def filter_records(records):
            shared_only = kwargs.get('shared')
            filtered = [r for r in records if not shared_only or r.shared]
            urls = kwargs.get('url')
            filtered = [r for r in filtered for url in urls if r.data.get('url') and url in r.data.get('url')] if urls \
                else filtered
            r_refs = kwargs.get('record')
            from fnmatch import fnmatch

            def title_match(title):
                return any([ref for ref in r_refs if fnmatch(title, ref)]) if title else False

            filtered = [r for r in filtered if r.record_uid in r_refs or title_match(r.data.get('title'))] if r_refs \
                else filtered
            deleted_items = kwargs.get('deleted_items')
            active_items = kwargs.get('active_items')
            if active_items and deleted_items:
                error_msg = '--deleted-items and --active-items flags are mutually exclusive'
                raise CommandError(self.get_parser().prog, error_msg)
            filtered = [r for r in filtered if r.in_trash] if deleted_items \
                else [r for r in filtered if not r.in_trash] if active_items \
                else filtered
            return filtered

        owners = filter_owners(sox_data.get_users().values())
        shared_folders = sox_data.get_shared_folders()
        permissions_lookup = dict()  # type: Dict[Tuple[str, str], str]

        def update_permissions_lookup(perm_lookup, rec_uid, enterprise_user, perm_bits):
            lookup_key = rec_uid, enterprise_user.email
            p_bits = perm_lookup.get(lookup_key, perm_bits)
            perm_lookup.update({lookup_key: perm_bits | p_bits})

        for owner in owners:
            owner_records = filter_records(sox_data.get_records(owner.records).values())
            for record in owner_records:
                if not record:
                    continue
                r_uid = record.record_uid
                for user_uid, permission_bits in record.user_permissions.items():
                    user = sox_data.get_user(user_uid)
                    update_permissions_lookup(permissions_lookup, r_uid, user, permission_bits)
                for folder in shared_folders.values():
                    for rp in folder.record_permissions:
                        if rp.record_uid == r_uid:
                            for user in sox_data.get_users(folder.users).values():
                                update_permissions_lookup(permissions_lookup, r_uid, user, rp.permission_bits)
                            for team in sox_data.get_teams(folder.teams).values():
                                for team_user in sox_data.get_users(team.users).values():
                                    update_permissions_lookup(permissions_lookup, r_uid, team_user, rp.permission_bits)

        table = []
        for key, permission_bits in permissions_lookup.items():
            r_uid, email = key
            table.append({'record_uid': r_uid, 'email': email, 'permissions': permission_bits})

        def format_table(rows):
            rows.sort(key=operator.itemgetter('permissions'), reverse=True)
            rows.sort(key=lambda item: item.get('permissions') & 1, reverse=True)
            rows.sort(key=operator.itemgetter('record_uid'))
            last_rec_uid = ''
            formatted_rows = []
            record_lookup = sox_data.get_records()
            for row in rows:
                rec_uid = row.get('record_uid')
                rec = record_lookup.get(rec_uid)
                r_data = rec.data
                r_title = r_data.get('title', '')
                r_type = r_data.get('record_type', '')
                r_url = r_data.get('url', '')
                rec_sfs = sox_data.get_record_sfs(rec_uid)
                formatted_rec_uid = rec_uid if report_fmt != 'table' or kwargs.get('pattern') or last_rec_uid != rec_uid else ''
                u_email = row.get('email')
                permissions = RecordPermissions.to_permissions_str(row.get('permissions'))
                fmt_row = [formatted_rec_uid, r_title, r_type, u_email, permissions, r_url.rstrip('/'), rec.in_trash, rec_sfs]
                formatted_rows.append(fmt_row)
                last_rec_uid = rec_uid
            return formatted_rows

        report_data = format_table(table)
        return report_data


class ComplianceTeamReportCommand(BaseComplianceReportCommand):
    def __init__(self):
        headers = ['team_name', 'team_uid', 'shared_folder_name', 'shared_folder_uid', 'permissions', 'records']
        super(ComplianceTeamReportCommand, self).__init__(headers, allow_no_opts=True)

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return team_report_parser

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        def get_sf_name(uid):
            # TODO: get sf name via share admin data when implemented (for now, find in params.shared_folder_cache)
            sf = params.shared_folder_cache.get(uid, {})
            return sf.get('name_unencrypted', '')

        show_team_users = kwargs.get('show_team_users')
        shared_folders = sox_data.get_shared_folders().items()
        team_lookup = sox_data.get_teams()
        report_data = []
        for sf_uid, folder in shared_folders:
            num_recs = len(folder.record_permissions) if folder.record_permissions else 0
            for team_uid in folder.teams:
                team = team_lookup.get(team_uid)
                perms = [not team.restrict_share and 'Can Share', not team.restrict_edit and 'Can Edit']
                perms = [p for p in perms if p]
                perms = ', '.join(perms) if perms else 'Read Only'
                row = [team.team_name, team_uid, get_sf_name(sf_uid), sf_uid, perms, num_recs]
                if show_team_users:
                    row.append(get_team_usernames(sox_data, team))
                report_data.append(row)

        return report_data

    def execute(self, params, **kwargs):  # type: (KeeperParams, any) -> any
        kwargs['shared'] = True
        tu_header = 'team_users'
        show_team_users = kwargs.get('show_team_users')
        if show_team_users:
            if tu_header not in self.report_headers:
                self.report_headers.append(tu_header)
        else:
            if tu_header in self.report_headers:
                self.report_headers.remove(tu_header)
        return super().execute(params, **kwargs)


class ComplianceRecordAccessReportCommand(BaseComplianceReportCommand):
    def __init__(self):
        super(ComplianceRecordAccessReportCommand, self).__init__([], allow_no_opts=True, prelim_only=False)
        self.group_by_column = 0

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return access_report_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, any) -> any
        kwargs['shared'] = True
        return super().execute(params, **kwargs)

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        # type: (KeeperParams, Dict[str, Any], sox.sox_data.SoxData, str, int, int) -> List[List[str]]
        def get_records_accessed_rq(email, filter_recs=None, created_max=None):
            # type: (str, Optional[List[str]], Optional[int]) -> Union[None, Dict[str, Any]]
            # Empty record filter list -> no records to search for
            if filter_recs is not None and not filter_recs:
                return None

            columns = ['record_uid', 'ip_address', 'keeper_version']
            rq_filter = {'username': email}
            if filter_recs is not None:
                rq_filter['record_uid'] = filter_recs
            if created_max:
                rq_filter['created'] = {'max': created_max}

            return dict(
                command         = 'get_audit_event_reports',
                report_type     = 'span',
                scope           = 'enterprise',
                aggregate       = ['last_created'],
                limit           = API_EVENT_SUMMARY_ROW_LIMIT,
                filter          = rq_filter,
                columns         = columns
            )

        # Extract data and meta-data from the server response, which determines the next request and/or its filter params
        # 2nd value of returned Tuple (filters for the next request) is None if there are no more events to fetch
        def process_access_events(events, filter_recs=None):
            records_accessed = dict()   # type: Dict[str, Dict[str, Any]]
            recs_set = set(filter_recs or [])

            for event in events:
                r_uid = event.get('record_uid')
                records_accessed.setdefault(r_uid, event)
                recs_set.discard(r_uid)

            queries_done = len(events) < API_EVENT_SUMMARY_ROW_LIMIT or not recs_set and filter_recs is not None
            filter_params = None
            if not queries_done:
                earliest_event = {} if not events else events[-1]
                filter_params = dict(
                    filter_recs =   list(recs_set) if filter_recs is not None else filter_recs,
                    created_max =   int(earliest_event.get('last_created', 0))
                )

            return records_accessed, filter_params

        def format_datetime(dt_str):
            if not dt_str:
                return None
            ts = datetime.datetime.fromisoformat(dt_str).timestamp()
            return datetime.datetime.fromtimestamp(int(ts))

        def from_ts(ts):
            return datetime.datetime.fromtimestamp(ts) if ts else None

        def compile_user_report(user, access_events):
            accessed_records = dict()
            rec_uids = access_events.keys() if report_type == report_type_default \
                else sox_data.get_vault_records(user).keys()

            for uid in rec_uids:
                access_event = access_events.get(uid, {})
                sox_rec = sox_data.get_records().get(uid)
                rec_info = sox_rec.data if sox_rec else {}
                rec_owner = sox_data.get_record_owner(uid)
                event_ts = access_event.get('last_created')
                accessed_record = {uid: {'record_title': rec_info.get('title'),
                                       'record_type':  rec_info.get('record_type'),
                                       'record_url': rec_info.get('url', '').rstrip('/'),
                                       'record_owner': rec_owner and rec_owner.email,
                                       'has_attachments': sox_rec.has_attachments if sox_rec else None,
                                       'in_trash': sox_rec.in_trash if sox_rec else None,
                                       'ip_address': access_event.get('ip_address'),
                                       'device': access_event.get('keeper_version'),
                                       'last_access': from_ts(int(event_ts)) if event_ts else None,
                                       'vault_owner': user}}
                accessed_records.update(accessed_record)
            return accessed_records

        def get_aging_data(rec_ids):
            if not rec_ids:
                return {}
            aging_data = {r: {'created': None, 'last_modified': None, 'last_rotation': None} for r in rec_ids}
            now = datetime.datetime.now()
            max_stored_age_dt = now - datetime.timedelta(days=1)
            max_stored_age_ts = int(max_stored_age_dt.timestamp())
            stored_entities = sox_data.storage.get_record_aging().get_all()
            stored_aging_data = {e.record_uid: {'created': from_ts(e.created), 'last_modified': from_ts(e.last_modified), 'last_rotation': from_ts(e.last_rotation)} for e in stored_entities}
            aging_data.update(stored_aging_data)

            def get_requests(filter_recs, filter_type, order='desc', aggregate='last_created'):
                columns = ['record_uid']
                requests = []
                while filter_recs:
                    chunk = filter_recs[:API_EVENT_SUMMARY_ROW_LIMIT]
                    filter_recs = filter_recs[API_EVENT_SUMMARY_ROW_LIMIT:]
                    request = dict(
                        command         = 'get_audit_event_reports',
                        report_type     = 'span',
                        scope           = 'enterprise',
                        aggregate       = [aggregate],
                        limit           = API_EVENT_SUMMARY_ROW_LIMIT,
                        filter          = dict(record_uid=chunk, audit_event_type=filter_type),
                        columns         = columns,
                        order           = order
                    )
                    requests.append(request)
                return requests

            def get_request_params(record_aging_event):
                # type: (str) -> Tuple[List[str], Union[List[str], None], Optional[str], Optional[str]]
                known_events_map = get_known_aging_data(record_aging_event)
                filter_recs = [uid for uid in rec_ids if uid not in known_events_map]
                types_by_aging_event = dict(
                    created         = None,
                    last_modified   = ['record_update'],
                    last_rotation   = ['record_rotation_scheduled_ok', 'record_rotation_on_demand_ok']
                )
                filter_types = types_by_aging_event.get(record_aging_event)
                order, aggregate = ('asc', 'first_created') if record_aging_event == 'created' \
                    else ('desc', 'last_created')
                return filter_recs, filter_types, order, aggregate

            def fetch_events(requests):
                return list(
                    itertools.chain.from_iterable(
                        [rs.get('audit_event_overview_report_rows', []) for rs in api.execute_batch(params, requests)]
                    )
                )

            def get_aging_events(aging_prop):
                req_params = get_request_params(aging_prop)
                requests = get_requests(*req_params)
                return fetch_events(requests)

            def get_known_aging_data(event_type):
                return {r: events.get(event_type) for r, events in stored_aging_data.items() if events.get(event_type) or 0 >= max_stored_age_ts}

            def get_aging_event_dts(event_type):
                events = get_aging_events(event_type)
                aggregate = 'first_created' if event_type == 'created' else 'last_created'
                record_timestamps = {event.get('record_uid', ''): event.get(aggregate) for event in events}
                return {rec: format_datetime(ts) for rec, ts in record_timestamps.items()}

            aging_stats = ['created', 'last_modified', 'last_rotation']
            record_events_by_stat = {stat: get_aging_event_dts(stat) for stat in aging_stats}
            for stat, record_event_dts in record_events_by_stat.items():
                for record, dt in record_event_dts.items():
                    aging_data.get(record, {}).update({stat: dt})
                    stat == 'created' and aging_data.get(record, {}).setdefault('last_modified', dt)

            save_aging_data(aging_data)
            return aging_data

        def save_aging_data(aging_data):
            existing_entities = sox_data.storage.get_record_aging()
            updated_entities = []
            for r, events in aging_data.items():
                entity = existing_entities.get_entity(r) or StorageRecordAging(r)
                created_dt = events.get('created')
                created_ts = int(created_dt.timestamp()) if created_dt else 0
                modified_dt = events.get('last_modified')
                modified_ts = int(modified_dt.timestamp()) if modified_dt else 0
                rotation_dt = events.get('last_rotation')
                rotation_ts = int(rotation_dt.timestamp()) if rotation_dt else 0

                entity.created = created_ts
                entity.last_modified = modified_ts
                entity.last_rotation = rotation_ts
                updated_entities.append(entity)
            sox_data.storage.record_aging.put_entities(updated_entities)

        def compile_report_data(rec_ids):
            aging_data = get_aging_data(rec_ids)
            for email, records in user_access_lookup.items():
                for uid, access_data in records.items():
                    row = [email, uid]
                    for i in range(len(row), len(self.report_headers)):
                        field = self.report_headers[i]
                        value = aging_data.get(uid, {}).get(field) if field in aging_columns \
                            else (access_data.get(field))
                        row.append(value)
                    report_data.append(row)

        def get_records_accessed(emails, limit_to_vault=False):
            # type: (List[str], Optional[bool]) -> Dict[str, Dict[str, List[str]]]
            get_rec_filter = lambda e: list(sox_data.get_vault_records(e).keys()) if limit_to_vault else None
            records_accessed_by_user = {e: dict() for e in emails}
            filters_by_user = {e: dict(filter_recs=get_rec_filter(e)) for e in emails}
            should_query = lambda rq_filter: rq_filter and (rq_filter.get('filter_recs') or not limit_to_vault)
            # Make requests in batches, walking backwards in time (w/ query filters) for all users in parallel (1 user per sub-request)
            while True:
                users_to_query = [user for user, user_filter in filters_by_user.items() if should_query(user_filter)]
                if not users_to_query:
                    break
                requests = [get_records_accessed_rq(email, **filters_by_user.get(email)) for email in users_to_query]
                responses = api.execute_batch(params, requests)
                responses_by_user = zip(users_to_query, responses)
                for user, response in responses_by_user:
                    access_events = response.get('audit_event_overview_report_rows', [])
                    records_accessed = records_accessed_by_user.get(user, {})
                    records_accessed_new, filters = process_access_events(access_events, filter_recs=filters_by_user.get(user, {}).get('filter_recs'))
                    for rec_uid, event in records_accessed_new.items():
                        records_accessed.setdefault(rec_uid, event)
                    records_accessed_by_user.update({user: records_accessed})
                    filters_by_user.update({user: filters})
            return records_accessed_by_user

        from ..sox.storage_types import StorageRecordAging
        from .aram import API_EVENT_SUMMARY_ROW_LIMIT
        from .enterprise import EnterpriseInfoCommand

        report_data = []
        user_lookup = {user.get('enterprise_user_id'): user.get('username') for user in params.enterprise.get('users')}
        user_access_lookup = dict()
        aging = kwargs.get('aging')
        users = kwargs.get('email') or ['@all']
        managed_users = json.loads( EnterpriseInfoCommand().execute(params, users=True, quiet=True, format='json'))
        usernames = [user_lookup.get(user_id) for user_id in sox_data.get_users()] if '@all' in users \
            else [user_lookup.get(int(ref)) if ref.isdigit() else ref for ref in users]
        usernames = [u for u in usernames if u and u in [mu.get('email') for mu in managed_users]]

        report_type_default = self.get_parser().get_default('report_type')
        report_type = kwargs.get('report_type', report_type_default)
        if report_type not in ACCESS_REPORT_TYPES:
            error_msg = f'Unrecognized report-type: "{report_type}"\nValues allowed: {ACCESS_REPORT_TYPES}'
            raise CommandError(self.get_parser().prog, error_msg)

        default_columns = ['vault_owner', 'record_uid', 'record_title', 'record_type', 'record_url', 'has_attachments',
                           'in_trash', 'record_owner', 'ip_address', 'device', 'last_access']

        aging_columns = ['created', 'last_modified', 'last_rotation'] if aging else []
        self.report_headers = default_columns + aging_columns

        record_access_events = get_records_accessed(usernames, report_type != report_type_default)
        user_access_lookup = {user: compile_user_report(user, access_events) for user, access_events in record_access_events.items()}
        record_ids = {r for recs in user_access_lookup.values() for r in recs} if aging else {}
        compile_report_data(record_ids)
        return report_data


class ComplianceSummaryReportCommand(BaseComplianceReportCommand):
    def __init__(self):
        headers = ['email', 'total_items', 'total_owned', 'active_owned', 'deleted_owned']
        super(ComplianceSummaryReportCommand, self).__init__(headers, allow_no_opts=True, prelim_only=False)

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return summary_report_parser

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        # type: (KeeperParams, any, SoxData, str, int, int) -> List[Iterable[any]]
        def get_row(u):
            num_deleted = len(u.trash_records)
            num_active = len(u.active_records)
            num_owned = len(u.records)
            vault_recs = sox_data.get_vault_records(u.user_uid)
            num_total = len(vault_recs)
            email = managed_user_email_lookup.get(u.user_uid) or u.email
            return email, num_total, num_owned, num_active, num_deleted

        filter_by_node = node != root_node
        from .enterprise import EnterpriseInfoCommand
        cmd = EnterpriseInfoCommand()
        cmd_kwargs = {
            'users': True,
            'quiet': True,
            'format': 'json',
            'node': str(node) if filter_by_node else None
        }
        cmd_rs = partial(cmd.execute, params, **cmd_kwargs)()
        managed_users = json.loads(cmd_rs)
        managed_users = [mu for mu in managed_users if mu.get('status', '').lower() != 'invited']
        managed_user_email_lookup = {mu.get('user_id'): mu.get('email') for mu in managed_users}
        managed_user_ids = set(managed_user_email_lookup.keys())
        empty_vault_user_ids = {user_id for user_id in managed_user_ids if user_id not in sox_data.get_users()}
        empty_vault_users = [mu for mu in managed_users if mu.get('user_id') in empty_vault_user_ids]
        sox_users = sox_data.get_users().values()
        report_data = [get_row(u) for u in sox_users if not filter_by_node or u.user_uid in managed_user_ids]
        report_data.extend([(u.get('email'), 0, 0, 0, 0) for u in empty_vault_users])
        total_active = sum([num_active for _, _, _, num_active, _ in report_data])
        total_deleted = sum([num_deleted for _, _, _, _, num_deleted in report_data])
        total_owned = sum([owned for _, _, owned, _, _ in report_data])
        report_data.append(('TOTAL', None, total_owned, total_active, total_deleted))
        return report_data


class ComplianceSharedFolderReportCommand(BaseComplianceReportCommand):
    def __init__(self):
        headers = ['shared_folder_uid', 'team_uid', 'team_name', 'record_uid', 'email']
        super(ComplianceSharedFolderReportCommand, self).__init__(headers, allow_no_opts=True)

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return sf_report_parser

    def generate_report_data(self, params, kwargs, sox_data, report_fmt, node, root_node):
        show_team_users = kwargs.get('show_team_users')
        team_users_title = '(TU) denotes a user whose membership in a team grants them access to the shared folder'
        self.title = team_users_title if show_team_users else None
        sfs = sox_data.get_shared_folders()
        teams = sox_data.get_teams()
        report_data = []
        for sfuid, sf in sfs.items():
            sf_team_uids = list(sf.teams)
            sf_team_names = [teams.get(t).team_name for t in sf.teams]
            records = [rp.record_uid for rp in sf.record_permissions]
            users = [get_email(sox_data, u) for u in sf.users]
            team_users = [tu for tuid in sf_team_uids for tu in get_team_usernames(sox_data, teams.get(tuid))] if show_team_users else []
            team_users = [f'(TU){email}' for email in team_users]
            row = [sfuid, sf_team_uids, sf_team_names, records, team_users + users]
            report_data.append(row)
        return report_data

